from __future__ import annotations

from pathlib import Path

from loguru import logger
from lxml import etree

from dsp_tools.error.exceptions import InputError
from dsp_tools.utils.xml_parsing.parse_xml import parse_xml_file
from dsp_tools.utils.xml_parsing.transform import remove_comments_from_element_tree
from dsp_tools.utils.xml_parsing.transform import transform_into_localnames
from dsp_tools.utils.xml_parsing.transform import transform_special_tags_make_localname
from dsp_tools.utils.xml_parsing.xml_schema_validation import validate_xml_against_schema


def parse_and_clean_xml_file(input_file: Path) -> etree._Element:
    """
    Parse an XML file with DSP-conform data,
    remove namespace URI from the elements' names,
    and transform the special tags `<region>`, `<link>`, `<video-segment>`, `<audio-segment>`
    to their technically correct form
    `<resource restype="Region">`, `<resource restype="LinkObj">`,
    `<resource restype="VideoSegment">`, `<resource restype="AudioSegment">`.

    Args:
        input_file: path to the XML file

    Returns:
        the root element of the parsed XML file

    Raises:
        InputError: if the input is not of either the expected types
    """
    root = parse_xml_file(input_file)
    root = remove_comments_from_element_tree(root)
    return transform_special_tags_make_localname(root)


def get_root_for_deserialisation(input_file: Path) -> etree._Element:
    root = parse_xml_file(input_file)
    root = remove_comments_from_element_tree(root)
    problem_msg = validate_xml_against_schema(root)
    if problem_msg:
        logger.error(problem_msg)
        raise InputError(problem_msg)
    print("The XML file is syntactically correct.")
    return transform_into_localnames(root)



from rdflib import Graph

from dsp_tools.commands.validate_data.validate_ontology import _reformat_ontology_validation_result


def test_reformat_ontology_validation_result() -> None:
    val_result = """
    @prefix api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#> .
    @prefix knora-api: <http://api.knora.org/ontology/knora-api/v2#> .
    @prefix sh: <http://www.w3.org/ns/shacl#> .
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
    
    [ a sh:ValidationResult ;
            sh:focusNode <http://0.0.0.0:3333/ontology/9991/error/v2#ImageWithKnoraProp_MissingSeqnum> ;
            sh:resultMessage "Message from the turtle file" ;
            sh:resultPath knora-api:isPartOf ;
            sh:resultSeverity sh:Violation ;
            sh:sourceConstraint _:n5badee8ae46f40828e18b000160a9e58b9 ;
            sh:sourceConstraintComponent sh:SPARQLConstraintComponent ;
            sh:sourceShape api-shapes:FindCardinalityMismatchSeqnum_OntologyShape ;
            sh:value <http://0.0.0.0:3333/ontology/9991/error/v2#ImageWithKnoraProp_MissingSeqnum> ] .
    """
    g = Graph()
    g.parse(data=val_result, format="ttl")
    reformatted = _reformat_ontology_validation_result(g)
    assert len(reformatted) == 1
    result = reformatted.pop(0)
    assert result.res_iri == "error:ImageWithKnoraProp_MissingSeqnum"
    assert result.msg == "Message from the turtle file"



# mypy: disable-error-code="method-assign,no-untyped-def"

import pytest

from dsp_tools.commands.validate_data.get_user_validation_message import _filter_out_duplicate_text_value_problem
from dsp_tools.commands.validate_data.get_user_validation_message import _get_message_for_one_resource
from dsp_tools.commands.validate_data.models.input_problems import InputProblem
from dsp_tools.commands.validate_data.models.input_problems import ProblemType


@pytest.fixture
def generic_problem() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.GENERIC,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="onto:hasGenericProblem",
        message="This is a generic problem.",
    )


@pytest.fixture
def file_value() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.FILE_VALUE,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="bitstream / iiif-uri",
        expected="A MovingImageRepresentation requires a file with the extension 'mp4'.",
    )


@pytest.fixture
def max_card() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.MAX_CARD,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="onto:hasMaxCardProblem",
        expected="Cardinality 1",
    )


@pytest.fixture
def min_card() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.MIN_CARD,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="onto:hasMinCardProblem",
        expected="Cardinality 1-n",
    )


@pytest.fixture
def non_existing_card() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.NON_EXISTING_CARD,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="onto:hasProp",
    )


@pytest.fixture
def file_value_prohibited() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.FILE_VALUE_PROHIBITED,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="bitstream / iiif-uri",
    )


@pytest.fixture
def link_value_type_mismatch() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.VALUE_TYPE_MISMATCH,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="onto:hasProp",
        input_type="LinkValue",
        expected="ListValue",
    )


@pytest.fixture
def input_regex() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.INPUT_REGEX,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="onto:hasProp",
        input_value="wrong input",
        expected="Expected format information",
    )


@pytest.fixture
def link_target_mismatch() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.LINK_TARGET_TYPE_MISMATCH,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="onto:hasProp",
        input_value="link_target_id",
        input_type="onto:Class",
        expected="onto:File or a subclass",
    )


@pytest.fixture
def inexistent_linked_resource() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.INEXISTENT_LINKED_RESOURCE,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="onto:hasProp",
        input_value="link_target_id",
    )


@pytest.fixture
def duplicate_value() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.DUPLICATE_VALUE,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="onto:hasProp",
        input_value="Text",
    )


def test_filter_out_duplicate_text_value_problem_with_duplicate(duplicate_value, link_value_type_mismatch):
    should_remain = InputProblem(
        problem_type=ProblemType.VALUE_TYPE_MISMATCH,
        res_id="should_remain",
        res_type="",
        prop_name="onto:hasProp",
        expected="TextValue without formatting",
    )
    should_be_removed = InputProblem(
        problem_type=ProblemType.VALUE_TYPE_MISMATCH,
        res_id="should_be_removed",
        res_type="",
        prop_name="onto:hasProp",
        expected="This property requires a TextValue",
    )
    result = _filter_out_duplicate_text_value_problem(
        [duplicate_value, link_value_type_mismatch, should_remain, should_be_removed]
    )
    assert len(result) == 3
    assert set([x.res_id for x in result]) == {"should_remain", "res_id"}


def test_filter_out_duplicate_text_value_problem_different_props():
    one = InputProblem(
        problem_type=ProblemType.VALUE_TYPE_MISMATCH,
        res_id="one",
        res_type="",
        prop_name="onto:prop2",
        expected="TextValue without formatting",
    )
    two = InputProblem(
        problem_type=ProblemType.VALUE_TYPE_MISMATCH,
        res_id="two",
        res_type="",
        prop_name="onto:prop1",
        expected="This property requires a TextValue",
    )
    result = _filter_out_duplicate_text_value_problem([one, two])
    assert len(result) == 2
    assert set([x.res_id for x in result]) == {"one", "two"}


def test_get_message_for_one_resource_generic(generic_problem):
    result = _get_message_for_one_resource([generic_problem])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\nonto:hasGenericProblem\n    - This is a generic problem."
    )
    assert result == expected


def test_get_message_for_one_resource_file_value(file_value):
    result = _get_message_for_one_resource([file_value])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "bitstream / iiif-uri\n"
        "    - Expected: A MovingImageRepresentation requires a file with the extension 'mp4'."
    )
    assert result == expected


def test_get_message_for_one_resource_max_card(max_card):
    result = _get_message_for_one_resource([max_card])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "onto:hasMaxCardProblem\n"
        "    - Maximum Cardinality Violation | Expected: Cardinality 1"
    )
    assert result == expected


def test_get_message_for_one_resource_min_card(min_card):
    result = _get_message_for_one_resource([min_card])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "onto:hasMinCardProblem\n"
        "    - Minimum Cardinality Violation | Expected: Cardinality 1-n"
    )
    assert result == expected


def test_get_message_for_one_resource_non_existing_card(non_existing_card):
    result = _get_message_for_one_resource([non_existing_card])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "onto:hasProp\n"
        "    - The resource class does not have a cardinality for this property."
    )
    assert result == expected


def test_get_message_for_one_resource_file_value_prohibited(file_value_prohibited):
    result = _get_message_for_one_resource([file_value_prohibited])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "bitstream / iiif-uri\n"
        "    - A file was added to the resource. This resource type must not have a file."
    )
    assert result == expected


def test_get_message_for_one_resource_value_type_mismatch(link_value_type_mismatch):
    result = _get_message_for_one_resource([link_value_type_mismatch])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "onto:hasProp\n"
        "    - Value Type Mismatch | Actual input type: 'LinkValue' | Expected Value Type: ListValue"
    )
    assert result == expected


def test_get_message_for_one_resource_input_regex(input_regex):
    result = _get_message_for_one_resource([input_regex])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "onto:hasProp\n"
        "    - Wrong Format of Input | Your input: 'wrong input' | "
        "Expected Input Format: Expected format information"
    )
    assert result == expected


def test_get_message_for_one_resource_link_target_mismatch(link_target_mismatch):
    result = _get_message_for_one_resource([link_target_mismatch])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "onto:hasProp\n"
        "    - Linked Resource Type Mismatch | Your input: 'link_target_id' | "
        "Actual input type: 'onto:Class' | Expected Resource Type: onto:File or a subclass"
    )
    assert result == expected


def test_get_message_for_one_resource_inexistent_linked_resource(inexistent_linked_resource):
    result = _get_message_for_one_resource([inexistent_linked_resource])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "onto:hasProp\n"
        "    - Linked Resource does not exist | Your input: 'link_target_id'"
    )
    assert result == expected


def test_get_message_for_one_resource_duplicate_value(duplicate_value):
    result = _get_message_for_one_resource([duplicate_value])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "onto:hasProp\n"
        "    - Your input is duplicated | Your input: 'Text'"
    )
    assert result == expected


def test_get_message_for_one_resource_several_problems(file_value, inexistent_linked_resource):
    result = _get_message_for_one_resource([file_value, inexistent_linked_resource])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "bitstream / iiif-uri\n"
        "    - Expected: A MovingImageRepresentation requires a file with the extension 'mp4'.\n"
        "onto:hasProp\n"
        "    - Linked Resource does not exist | Your input: 'link_target_id'"
    )
    assert result == expected



import pytest
from rdflib import RDF
from rdflib import SH
from rdflib import BNode
from rdflib import Graph
from rdflib import Literal
from rdflib import URIRef

from dsp_tools.commands.validate_data.sparql.cardinality_shacl import _construct_0_1_cardinality
from dsp_tools.commands.validate_data.sparql.cardinality_shacl import _construct_0_n_cardinality
from dsp_tools.commands.validate_data.sparql.cardinality_shacl import _construct_1_cardinality
from dsp_tools.commands.validate_data.sparql.cardinality_shacl import _construct_1_n_cardinality
from dsp_tools.commands.validate_data.sparql.cardinality_shacl import _construct_all_cardinalities
from dsp_tools.commands.validate_data.sparql.cardinality_shacl import _construct_resource_nodeshape
from dsp_tools.commands.validate_data.sparql.cardinality_shacl import construct_cardinality_node_shapes
from dsp_tools.utils.rdflib_constants import DASH
from test.unittests.commands.validate_data.constants import ONTO


class TestCheckTripleNumbersOnto:
    def test_nodeshape(self, onto_for_cardinality: Graph) -> None:
        result = _construct_resource_nodeshape(onto_for_cardinality)
        number_of_resource_classes = 12
        triples_cls_nodeshape = 13 * number_of_resource_classes
        assert len(result) == triples_cls_nodeshape

    def test_cardinality_1(self, onto_for_cardinality: Graph) -> None:
        result = _construct_1_cardinality(onto_for_cardinality)
        number_of_occurrences_in_onto = 2
        triples_card_1 = 7 * number_of_occurrences_in_onto
        assert len(result) == triples_card_1

    def test_cardinality_0_1(self, onto_for_cardinality: Graph) -> None:
        result = _construct_0_1_cardinality(onto_for_cardinality)
        number_of_occurrences_in_onto = 5
        triples_card_0_1 = 7 * number_of_occurrences_in_onto
        assert len(result) == triples_card_0_1

    def test_cardinality_0_n(self, onto_for_cardinality: Graph) -> None:
        result = _construct_0_n_cardinality(onto_for_cardinality)
        number_of_occurrences_in_onto = 22  # Inheritance included
        triples_card_0_n = 3 * number_of_occurrences_in_onto
        assert len(result) == triples_card_0_n

    def test_cardinality_1_n(self, onto_for_cardinality: Graph) -> None:
        result = _construct_1_n_cardinality(onto_for_cardinality)
        number_of_occurrences_in_onto = 1
        triples_card_1_n = 6 * number_of_occurrences_in_onto
        assert len(result) == triples_card_1_n


def test_construct_cardinality_node_shapes(onto_for_cardinality: Graph) -> None:
    result = construct_cardinality_node_shapes(onto_for_cardinality)
    expected_classes = {
        ONTO.ClassInheritedCardinality,
        ONTO.ClassMixedCard,
        ONTO.CardOneResource,
        ONTO.ClassWithEverything,
        ONTO.TestStillImageRepresentation,
        ONTO.ClassInheritedCardinalityOverwriting,
        ONTO.TestArchiveRepresentation,
        ONTO.TestAudioRepresentation,
        ONTO.TestDocumentRepresentation,
        ONTO.TestMovingImageRepresentation,
        ONTO.TestStillImageRepresentation,
        ONTO.TestTextRepresentation,
        URIRef("http://0.0.0.0:3333/ontology/9999/second-onto/v2#SecondOntoClass"),
    }
    result_classes = set(result.subjects(DASH.closedByTypes, Literal(True)))
    assert result_classes == expected_classes


def test_construct_resource_nodeshape_one_res(one_res_one_prop: Graph) -> None:
    result = _construct_resource_nodeshape(one_res_one_prop)
    subjects = {iri for x in result.triples((None, None, None)) if not isinstance(iri := x[0], BNode)}
    assert len(subjects) == 1
    subject_iri = subjects.pop()
    assert subject_iri == ONTO.CardOneResource
    node_triples = list(result.triples((subject_iri, None, None)))
    num_triples = 5
    assert len(node_triples) == num_triples
    assert next(result.subjects(RDF.type, SH.NodeShape)) == subject_iri
    assert next(result.objects(subject_iri, DASH.closedByTypes)) == Literal(True)
    assert isinstance(next(result.objects(subject_iri, SH.property)), BNode)


def test_construct_resource_nodeshape_no_res(one_bool_prop: Graph) -> None:
    result = _construct_resource_nodeshape(one_bool_prop)
    assert len(result) == 0


class Test1:
    def test_good(self, card_1: Graph) -> None:
        result = _construct_1_cardinality(card_1)
        assert len(result) == 7
        bn = next(result.subjects(RDF.type, SH.PropertyShape))
        shape_iri = next(result.subjects(SH.property, bn))
        assert shape_iri == ONTO.ClassMixedCard
        assert str(next(result.objects(bn, SH.minCount))) == "1"
        assert str(next(result.objects(bn, SH.maxCount))) == "1"
        assert next(result.objects(bn, SH.path)) == ONTO.testBoolean
        assert next(result.objects(bn, SH.severity)) == SH.Violation
        assert str(next(result.objects(bn, SH.message))) == "Cardinality 1"

    def test_good_link_value(self, link_prop_card_1: Graph) -> None:
        result = _construct_1_cardinality(link_prop_card_1)
        assert len(result) == 7
        bn = next(result.subjects(RDF.type, SH.PropertyShape))
        shape_iri = next(result.subjects(SH.property, bn))
        assert shape_iri == ONTO.ClassMixedCard
        assert str(next(result.objects(bn, SH.minCount))) == "1"
        assert str(next(result.objects(bn, SH.maxCount))) == "1"
        assert next(result.objects(bn, SH.path)) == ONTO.testHasLinkToCardOneResource
        assert next(result.objects(bn, SH.severity)) == SH.Violation
        assert str(next(result.objects(bn, SH.message))) == "Cardinality 1"

    def test_empty_0_1(self, card_0_1: Graph) -> None:
        result = _construct_1_cardinality(card_0_1)
        assert len(result) == 0

    def test_empty_1_n(self, card_1_n: Graph) -> None:
        result = _construct_1_cardinality(card_1_n)
        assert len(result) == 0

    def test_empty_0_n(self, card_0_n: Graph) -> None:
        result = _construct_1_cardinality(card_0_n)
        assert len(result) == 0


class Test01:
    def test_good(self, card_0_1: Graph) -> None:
        result = _construct_0_1_cardinality(card_0_1)
        assert len(result) == 7
        bn = next(result.subjects(RDF.type, SH.PropertyShape))
        shape_iri = next(result.subjects(SH.property, bn))
        assert shape_iri == ONTO.ClassMixedCard
        assert str(next(result.objects(bn, SH.minCount))) == "0"
        assert str(next(result.objects(bn, SH.maxCount))) == "1"
        assert next(result.objects(bn, SH.path)) == ONTO.testDecimalSimpleText
        assert next(result.objects(bn, SH.severity)) == SH.Violation
        assert str(next(result.objects(bn, SH.message))) == "Cardinality 0-1"

    def test_good_link_value(self, link_prop_card_01: Graph) -> None:
        result = _construct_0_1_cardinality(link_prop_card_01)
        assert len(result) == 7
        bn = next(result.subjects(RDF.type, SH.PropertyShape))
        shape_iri = next(result.subjects(SH.property, bn))
        assert shape_iri == ONTO.ClassMixedCard
        assert str(next(result.objects(bn, SH.minCount))) == "0"
        assert str(next(result.objects(bn, SH.maxCount))) == "1"
        assert next(result.objects(bn, SH.path)) == ONTO.testHasLinkToCardOneResource
        assert next(result.objects(bn, SH.severity)) == SH.Violation
        assert str(next(result.objects(bn, SH.message))) == "Cardinality 0-1"

    def test_empty_1(self, card_1: Graph) -> None:
        result = _construct_0_1_cardinality(card_1)
        assert len(result) == 0

    def test_empty_1_n(self, card_1_n: Graph) -> None:
        result = _construct_0_1_cardinality(card_1_n)
        assert len(result) == 0

    def test_empty_0_n(self, card_0_n: Graph) -> None:
        result = _construct_0_1_cardinality(card_0_n)
        assert len(result) == 0


class Test1N:
    def test_good(self, card_1_n: Graph) -> None:
        result = _construct_1_n_cardinality(card_1_n)
        assert len(result) == 6
        bn = next(result.subjects(RDF.type, SH.PropertyShape))
        shape_iri = next(result.subjects(SH.property, bn))
        assert shape_iri == ONTO.ClassMixedCard
        assert str(next(result.objects(bn, SH.minCount))) == "1"
        assert next(result.objects(bn, SH.path)) == ONTO.testGeoname
        assert next(result.objects(bn, SH.severity)) == SH.Violation
        assert str(next(result.objects(bn, SH.message))) == "Cardinality 1-n"

    def test_empty_1(self, card_1: Graph) -> None:
        result = _construct_1_n_cardinality(card_1)
        assert len(result) == 0

    def test_empty_1_n(self, card_0_1: Graph) -> None:
        result = _construct_1_n_cardinality(card_0_1)
        assert len(result) == 0

    def test_empty_0_n(self, card_0_n: Graph) -> None:
        result = _construct_1_n_cardinality(card_0_n)
        assert len(result) == 0


class Test0N:
    def test_good(self, card_0_n: Graph) -> None:
        result = _construct_0_n_cardinality(card_0_n)
        assert len(result) == 3
        bn = next(result.subjects(RDF.type, SH.PropertyShape))
        shape_iri = next(result.subjects(SH.property, bn))
        assert shape_iri == ONTO.ClassMixedCard
        assert next(result.objects(bn, SH.path)) == ONTO.testSimpleText

    def test_empty_1_n(self, card_1_n: Graph) -> None:
        result = _construct_0_n_cardinality(card_1_n)
        assert len(result) == 0

    def test_empty_1(self, card_1: Graph) -> None:
        result = _construct_0_n_cardinality(card_1)
        assert len(result) == 0

    def test_empty_0_1(self, card_0_1: Graph) -> None:
        result = _construct_0_n_cardinality(card_0_1)
        assert len(result) == 0


def test_construct_all_cardinalities(one_res_one_prop: Graph) -> None:
    result = _construct_all_cardinalities(one_res_one_prop)
    assert len(result) == 7
    bn = next(result.subjects(RDF.type, SH.PropertyShape))
    shape_iri = next(result.subjects(SH.property, bn))
    assert shape_iri == ONTO.CardOneResource
    assert str(next(result.objects(bn, SH.minCount))) == "1"
    assert str(next(result.objects(bn, SH.maxCount))) == "1"
    assert next(result.objects(bn, SH.path)) == ONTO.testBoolean
    assert next(result.objects(bn, SH.severity)) == SH.Violation
    assert str(next(result.objects(bn, SH.message))) == "Cardinality 1"


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from dataclasses import field
from typing import Protocol

from dsp_tools.commands.excel2json.models.input_error import ExcelFileProblem
from dsp_tools.commands.excel2json.models.input_error import PositionInExcel
from dsp_tools.commands.excel2json.models.input_error import grand_separator
from dsp_tools.commands.excel2json.models.input_error import list_separator
from dsp_tools.commands.excel2json.models.input_error import medium_separator
from dsp_tools.error.problems import Problem


@dataclass(frozen=True)
class ListCreationProblem:
    excel_problems: list[Problem]

    def execute_error_protocol(self) -> str:
        title = "\nThe excel file(s) used to create the list section have the following problem(s):\n\n"
        msg = [problem.execute_error_protocol() for problem in self.excel_problems]
        return title + grand_separator.join(msg)


@dataclass(frozen=True)
class ListNodeProblem:
    node_id: str
    problems: dict[str, str]

    def execute_error_protocol(self) -> str:
        msg = [f"The node '{self.node_id}' has the following problem(s):"]
        msg.extend([f"Field: '{key}', Problem: {value}" for key, value in self.problems.items()])
        return list_separator.join(msg)


@dataclass
class SheetProblem(Protocol):
    excel_name: str

    def execute_error_protocol(self) -> str:
        raise NotImplementedError


@dataclass
class CollectedSheetProblems:
    sheet_problems: list[SheetProblem]

    def execute_error_protocol(self) -> str:
        file_problems = self._group_and_sort_by_excel_name()
        return grand_separator.join([x.execute_error_protocol() for x in file_problems])

    def _group_and_sort_by_excel_name(self) -> list[ExcelFileProblem]:
        file_dict: dict[str, list[Problem]] = defaultdict(list)
        for problem in self.sheet_problems:
            file_dict[problem.excel_name].append(problem)
        file_list = [ExcelFileProblem(k, v) for k, v in file_dict.items()]
        return sorted(file_list, key=lambda x: x.filename)


@dataclass
class DuplicatesInSheetProblem(SheetProblem):
    excel_name: str
    sheet_name: str
    rows: list[int]

    def execute_error_protocol(self) -> str:
        msg = [
            f"The excel sheet '{self.sheet_name}' contains rows that are completely identical "
            f"(excluding the column 'ID (optional)'). The following rows are duplicates:"
        ]
        msg.extend([f"{x + 2}" for x in self.rows])
        return list_separator.join(msg)


@dataclass
class MultipleListPerSheetProblem(SheetProblem):
    excel_name: str
    sheet_name: str
    list_names: list[str]

    def execute_error_protocol(self) -> str:
        return (
            f"Per Excel sheet only one list is allowed.\n"
            f"The sheet '{self.sheet_name}' has more than one list, namely the following: {', '.join(self.list_names)}"
        )


@dataclass
class ListSheetComplianceProblem(SheetProblem):
    excel_name: str
    sheet_name: str
    problems: list[Problem]

    def execute_error_protocol(self) -> str:
        msg = [f"The excel sheet '{self.sheet_name}' has the following problem(s):"]
        msg.extend([x.execute_error_protocol() for x in self.problems])
        return list_separator.join(msg)


@dataclass
class ListSheetContentProblem(SheetProblem):
    excel_name: str
    sheet_name: str
    problems: list[Problem]

    def execute_error_protocol(self) -> str:
        msg = [f"The Excel sheet '{self.sheet_name}' has the following problem(s):"]
        msg.extend([problem.execute_error_protocol() for problem in self.problems])
        return list_separator.join(msg)


@dataclass
class MissingTranslationsSheetProblem(SheetProblem):
    excel_name: str
    sheet: str
    node_problems: list[MissingNodeTranslationProblem]

    def execute_error_protocol(self) -> str:
        msg = [
            f"The excel sheet '{self.sheet}' has the following problem(s):\n"
            "In one list, all the nodes must be translated into all the languages used. "
            "For the following nodes, the translations are missing:"
        ]
        nodes_sorted = sorted(self.node_problems, key=lambda x: x.index_num)
        msg.extend([x.execute_error_protocol() for x in nodes_sorted])
        return list_separator.join(msg)


@dataclass(frozen=True)
class DuplicatesListNameProblem:
    all_duplicate_names: list[ListInformation]

    def execute_error_protocol(self) -> str:
        msg = [
            "The name of the list must be unique across all the excel sheets.\n"
            "The following sheets have lists with the same name:"
        ]
        sorted_list = sorted(self.all_duplicate_names, key=lambda x: x.excel_name)
        msg.extend([x.execute_error_protocol() for x in sorted_list])
        return list_separator.join(msg)


@dataclass(frozen=True)
class ListInformation:
    excel_name: str
    excel_sheet: str
    list_name: str

    def execute_error_protocol(self) -> str:
        return f"Excel file: '{self.excel_name}', Sheet: '{self.excel_sheet}', List: '{self.list_name}'"


@dataclass(frozen=True)
class DuplicatesCustomIDInProblem:
    duplicate_ids: list[DuplicateIDProblem]

    def execute_error_protocol(self) -> str:
        msg = [
            "No duplicates are allowed in the 'ID (optional)' column. At the following locations, IDs are duplicated:"
        ]
        sorted_ids = sorted(self.duplicate_ids, key=lambda x: x.custom_id)
        msg.extend([x.execute_error_protocol() for x in sorted_ids])
        return medium_separator.join(msg)


@dataclass
class DuplicateIDProblem:
    custom_id: str | int | float = field(default="")
    excel_locations: list[PositionInExcel] = field(default_factory=list)

    def execute_error_protocol(self) -> str:
        msg = [f"ID: '{self.custom_id}'"]
        msg.extend([str(x) for x in sorted(self.excel_locations, key=lambda x: str(x.excel_filename))])
        return list_separator.join(msg)


@dataclass(frozen=True)
class MissingNodeTranslationProblem:
    empty_columns: list[str]
    index_num: int

    def execute_error_protocol(self) -> str:
        return f"Row Number: {self.index_num + 2} | Column(s): {', '.join(sorted(self.empty_columns))}"


@dataclass(frozen=True)
class NodesPerRowProblem:
    column_names: list[str]
    index_num: int
    should_be_empty: bool

    def execute_error_protocol(self) -> str:
        if self.should_be_empty:
            description = "Column(s) that must be empty"
        else:
            description = "Column(s) that must be filled"
        return f"Row Number: {self.index_num + 2}, {description}: {', '.join(self.column_names)}"


@dataclass(frozen=True)
class MinimumRowsProblem:
    def execute_error_protocol(self) -> str:
        return (
            "The Excel sheet must contain at least two rows, "
            "one for the list name and one row for a minimum of one node."
        )


@dataclass(frozen=True)
class MissingNodeColumn:
    def execute_error_protocol(self) -> str:
        return (
            "At least one column for the node names in the format '[lang]_[column_number]' is required. "
            "The allowed language tags are: en, de, fr, it, rm."
        )


@dataclass
class MissingExpectedColumn:
    missing_cols: set[str]

    def execute_error_protocol(self) -> str:
        return (
            f"All nodes and lists must be translated into the same languages. "
            f"Based on the languages used, the following column(s) are missing: {', '.join(self.missing_cols)}"
        )



import pytest
from lxml import etree

from dsp_tools.commands.xmlupload.prepare_xml_input.read_validate_xml_file import _check_if_resptr_targets_exist
from dsp_tools.commands.xmlupload.prepare_xml_input.read_validate_xml_file import _check_if_salsah_targets_exist


def test_check_if_resptr_targets_exist() -> None:
    """Check correct input"""
    xml = """
    <knora>
        <resource id="resource1">
            <resptr-prop name="resptr1"><resptr>resource2</resptr></resptr-prop>
        </resource>
        <resource id="resource2">
            <resptr-prop name="resptr2"><resptr>resource1</resptr></resptr-prop>
        </resource>
    </knora>
    """
    root = etree.fromstring(xml)
    errors_returned = _check_if_resptr_targets_exist(root)
    assert not errors_returned


def test_check_if_resptr_targets_exist_invalid() -> None:
    """Check invalid input"""
    xml = """
    <knora>
        <resource id="resource1">
            <resptr-prop name="resptr1"><resptr>resource3</resptr></resptr-prop>
        </resource>
        <resource id="resource2">
            <resptr-prop name="resptr2"><resptr>resource4</resptr></resptr-prop>
        </resource>
    </knora>
    """
    root = etree.fromstring(xml)
    errors_returned = _check_if_resptr_targets_exist(root)
    errors_expected = [
        "Resource 'resource1', property 'resptr1' has an invalid link target 'resource3'",
        "Resource 'resource2', property 'resptr2' has an invalid link target 'resource4'",
    ]
    assert errors_returned == errors_expected


def test_check_if_salsah_targets_exist() -> None:
    """Check correct input"""
    xml = """
    <knora>
        <resource id="resource1">
            <text-prop name="text1">
                <text encoding="xml">
                    <a class="salsah-link" href="IRI:resource2:IRI">resource2</a>
                </text>
            </text-prop>
        </resource>
        <resource id="resource2">
            <text-prop name="text2">
                <text encoding="xml">
                    <a class="salsah-link" href="IRI:resource1:IRI">resource1</a>
                </text>
            </text-prop>
        </resource>
    </knora>
    """
    root = etree.fromstring(xml)
    errors_returned = _check_if_salsah_targets_exist(root)
    assert not errors_returned


def test_check_if_salsah_targets_exist_invalid() -> None:
    """Check invalid input"""
    xml = """
    <knora>
        <resource id="resource1">
            <text-prop name="text1">
                <text encoding="xml">
                    <a class="salsah-link" href="IRI:resource3:IRI">resource3</a>
                </text>
            </text-prop>
        </resource>
        <resource id="resource2">
            <text-prop name="text2">
                <text encoding="xml">
                    <a class="salsah-link" href="IRI:resource4:IRI">resource4</a>
                </text>
            </text-prop>
        </resource>
    </knora>
    """
    root = etree.fromstring(xml)
    errors_returned = _check_if_salsah_targets_exist(root)
    errors_expected = [
        "Resource 'resource1', property 'text1' has an invalid link target 'IRI:resource3:IRI'",
        "Resource 'resource2', property 'text2' has an invalid link target 'IRI:resource4:IRI'",
    ]
    assert errors_returned == errors_expected


if __name__ == "__main__":
    pytest.main([__file__])



from dataclasses import dataclass
from typing import Any
from typing import cast
from urllib.parse import quote_plus

import requests
from rdflib import SH
from rdflib import Graph

from dsp_tools.commands.validate_data.models.api_responses import AllProjectLists
from dsp_tools.commands.validate_data.models.api_responses import OneList
from dsp_tools.commands.validate_data.models.api_responses import SHACLValidationReport
from dsp_tools.commands.validate_data.models.validation import RDFGraphs
from dsp_tools.error.exceptions import InternalError
from dsp_tools.utils.request_utils import PostFile
from dsp_tools.utils.request_utils import PostFiles
from dsp_tools.utils.request_utils import RequestParameters
from dsp_tools.utils.request_utils import log_request
from dsp_tools.utils.request_utils import log_response


@dataclass
class OntologyClient:
    api_url: str
    shortcode: str

    def get_knora_api(self) -> str:
        url = f"{self.api_url}/ontology/knora-api/v2#"
        headers = {"Accept": "text/turtle"}
        timeout = 60
        log_request(RequestParameters("GET", url, timeout=timeout, headers=headers))
        response = requests.get(url=url, headers=headers, timeout=timeout)
        log_response(response, include_response_content=False)
        if not response.ok:
            raise InternalError(f"Failed Request: {response.status_code} {response.text}")
        return response.text

    def get_ontologies(self) -> list[str]:
        """
        Returns a list of project ontologies as a string in turtle format.

        Returns:
            list of ontologies
        """
        ontology_iris = self._get_ontology_iris()
        return [self._get_one_ontology(x) for x in ontology_iris]

    def _get_ontology_iris(self) -> list[str]:
        url = f"{self.api_url}/admin/projects/shortcode/{self.shortcode}"
        timeout = 10
        log_request(RequestParameters("GET", url, timeout=timeout))
        response = requests.get(url=url, timeout=timeout)
        log_response(response)
        if not response.ok:
            raise InternalError(f"Failed Request: {response.status_code} {response.text}")
        response_json = cast(dict[str, Any], response.json())
        if not (ontos := response_json.get("project", {}).get("ontologies")):
            raise InternalError(f"The response from the API does not contain any ontologies.\nResponse:{response.text}")
        output = cast(list[str], ontos)
        return output

    def _get_one_ontology(self, ontology_iri: str) -> str:
        url = ontology_iri
        headers = {"Accept": "text/turtle"}
        timeout = 10
        log_request(RequestParameters("GET", url, timeout=timeout, headers=headers))
        response = requests.get(url=url, headers=headers, timeout=timeout)
        log_response(response, include_response_content=False)
        if not response.ok:
            raise InternalError(f"Failed Request: {response.status_code} {response.text}")
        return response.text


@dataclass
class ListClient:
    """Client to request and reformat the lists of a project."""

    api_url: str
    shortcode: str

    def get_lists(self) -> AllProjectLists:
        list_json = self._get_all_list_iris()
        all_iris = self._extract_list_iris(list_json)
        all_lists = [self._get_one_list(iri) for iri in all_iris]
        reformatted = [self._reformat_one_list(lst) for lst in all_lists]
        return AllProjectLists(reformatted)

    def _get_all_list_iris(self) -> dict[str, Any]:
        url = f"{self.api_url}/admin/lists?projectShortcode={self.shortcode}"
        timeout = 10
        log_request(RequestParameters("GET", url, timeout))
        response = requests.get(url=url, timeout=timeout)
        log_response(response)
        if not response.ok:
            raise InternalError(f"Failed Request: {response.status_code} {response.text}")
        json_response = cast(dict[str, Any], response.json())
        return json_response

    def _extract_list_iris(self, response_json: dict[str, Any]) -> list[str]:
        return [x["id"] for x in response_json["lists"]]

    def _get_one_list(self, list_iri: str) -> dict[str, Any]:
        encoded_list_iri = quote_plus(list_iri)
        url = f"{self.api_url}/admin/lists/{encoded_list_iri}"
        timeout = 10
        log_request(RequestParameters("GET", url, timeout))
        response = requests.get(url=url, timeout=timeout)
        log_response(response)
        if not response.ok:
            raise InternalError(f"Failed Request: {response.status_code} {response.text}")
        response_json = cast(dict[str, Any], response.json())
        return response_json

    def _reformat_one_list(self, response_json: dict[str, Any]) -> OneList:
        list_name = response_json["list"]["listinfo"]["name"]
        list_id = response_json["list"]["listinfo"]["id"]
        nodes = response_json["list"]["children"]
        all_nodes = []
        for child in nodes:
            all_nodes.append(child["name"])
            if node_child := child.get("children"):
                all_nodes.extend(self._reformat_children(node_child, all_nodes))
        return OneList(list_iri=list_id, list_name=list_name, nodes=all_nodes)

    def _reformat_children(self, list_child: list[dict[str, Any]], current_nodes: list[str]) -> list[str]:
        for child in list_child:
            current_nodes.append(child["name"])
            if grand_child := child.get("children"):
                self._reformat_children(grand_child, current_nodes)
        return current_nodes


@dataclass
class ShaclValidator:
    """Client to validate RDF data against a given SHACL shape."""

    api_url: str

    def validate(self, rdf_graphs: RDFGraphs) -> SHACLValidationReport:
        """
        Sends a multipart/form-data request with two turtle files (data.ttl and shacl.ttl) to the given URL
        and expects a response containing a single text/turtle body which is loaded into an rdflib Graph.

        Args:
            rdf_graphs: Graphs in RDF form that should be validated

        Returns:
            SHACLValidationReport: A report containing the validation graph and a bool to indicate if it passed.

        Raises:
            InternalError: in case of a non-ok response
        """
        result_graph = Graph()
        conforms = True

        card_result = self._validate_cardinality(rdf_graphs)
        if not card_result.conforms:
            result_graph += card_result.validation_graph
            conforms = False

        content_result = self._validate_content(rdf_graphs)
        if not content_result.conforms:
            result_graph += content_result.validation_graph
            conforms = False

        return SHACLValidationReport(conforms=conforms, validation_graph=result_graph)

    def validate_ontology(self, onto_graph: Graph, onto_shacl: Graph) -> SHACLValidationReport:
        post_files = self._prepare_validation_files_for_request(onto_graph, onto_shacl)
        url = f"{self.api_url}/shacl/validate"
        timeout = 60
        log_request(RequestParameters("POST", url, timeout, files=post_files))
        response = requests.post(url=url, files=post_files.to_dict(), timeout=timeout)
        log_response(response)
        if not response.ok:
            raise InternalError(f"Failed Request: {response.status_code} {response.text}")
        return self._parse_validation_result(response.text)

    def _validate_cardinality(self, rdf_graphs: RDFGraphs) -> SHACLValidationReport:
        card_files = self._prepare_cardinality_files(rdf_graphs)
        url = f"{self.api_url}/shacl/validate"
        timeout = 60
        log_request(RequestParameters("POST", url, timeout, files=card_files))
        response = requests.post(url=url, files=card_files.to_dict(), timeout=timeout)
        log_response(response)
        if not response.ok:
            raise InternalError(f"Failed Request: {response.status_code} {response.text}")
        return self._parse_validation_result(response.text)

    def _prepare_cardinality_files(self, rdf_graphs: RDFGraphs) -> PostFiles:
        shacl_graph = rdf_graphs.cardinality_shapes + rdf_graphs.ontos + rdf_graphs.knora_api
        return self._prepare_validation_files_for_request(rdf_graphs.data, shacl_graph)

    def _validate_content(self, rdf_graphs: RDFGraphs) -> SHACLValidationReport:
        content_files = self._prepare_content_files(rdf_graphs)
        url = f"{self.api_url}/shacl/validate"
        timeout = 60
        log_request(RequestParameters("POST", url, timeout, files=content_files))
        response = requests.post(url=url, files=content_files.to_dict(), timeout=timeout)
        log_response(response)
        if not response.ok:
            raise InternalError(f"Failed Request: {response.status_code} {response.text}")
        return self._parse_validation_result(response.text)

    def _prepare_content_files(self, rdf_graphs: RDFGraphs) -> PostFiles:
        shacl_graph = rdf_graphs.content_shapes + rdf_graphs.ontos + rdf_graphs.knora_api
        data_graph = rdf_graphs.data + rdf_graphs.ontos + rdf_graphs.knora_api
        return self._prepare_validation_files_for_request(data_graph, shacl_graph)

    @staticmethod
    def _prepare_validation_files_for_request(data_graph: Graph, shacl_graph: Graph) -> PostFiles:
        shacl_str = shacl_graph.serialize(format="ttl")
        shacl_file = PostFile(file_name="shacl.ttl", fileobj=shacl_str, content_type="text/turtle")
        data_str = data_graph.serialize(format="ttl")
        data_file = PostFile(file_name="data.ttl", fileobj=data_str, content_type="text/turtle")
        return PostFiles([shacl_file, data_file])

    def _parse_validation_result(self, response_text: str) -> SHACLValidationReport:
        graph = Graph()
        graph.parse(data=response_text, format="turtle")
        conforms = bool(next(graph.objects(None, SH.conforms)))
        return SHACLValidationReport(conforms=conforms, validation_graph=graph)



from typing import TypeAlias
from typing import Union

from rdflib import RDF
from rdflib import XSD
from rdflib import BNode
from rdflib import Graph
from rdflib import Literal
from rdflib import URIRef

from dsp_tools.commands.xmlupload.iri_resolver import IriResolver
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import LINK_PROP_TYPE_INFO
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import LIST_PROP_TYPE_INFO
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import RDF_LITERAL_PROP_TYPE_MAPPER
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import RICHTEXT_PROP_TYPE_INFO
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryBoolean
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryColor
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryDate
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryDecimal
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryGeometry
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryGeoname
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryInt
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryInterval
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryLink
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryList
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryRichtext
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediarySimpleText
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryTime
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryUri
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryValue
from dsp_tools.commands.xmlupload.models.lookup_models import IRILookups
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.models.rdf_models import RDFPropTypeInfo
from dsp_tools.config.logger_config import WARNINGS_SAVEPATH
from dsp_tools.error.exceptions import BaseError
from dsp_tools.error.exceptions import InputError
from dsp_tools.utils.data_formats.date_util import DayMonthYearEra
from dsp_tools.utils.data_formats.date_util import SingleDate
from dsp_tools.utils.data_formats.date_util import StartEnd
from dsp_tools.utils.data_formats.iri_util import is_resource_iri
from dsp_tools.utils.rdflib_constants import KNORA_API

LiteralValueTypesAlias: TypeAlias = Union[
    IntermediaryBoolean,
    IntermediaryColor,
    IntermediaryDecimal,
    IntermediaryGeoname,
    IntermediaryGeometry,
    IntermediaryInt,
    IntermediarySimpleText,
    IntermediaryTime,
    IntermediaryUri,
]


def make_values(values: list[IntermediaryValue], res_node: BNode | URIRef, lookups: IRILookups) -> Graph:
    """
    Serialise the values of a resource.

    Args:
        values: list of IntermediaryValues of the resource
        res_node: node of the resource
        lookups: lookups to resolve IRIs

    Returns:
        Graph with the values
    """
    properties_graph = Graph()
    for val in values:
        single_prop_graph = _make_one_value_graph(val=val, res_node=res_node, iri_lookups=lookups)
        properties_graph += single_prop_graph
    return properties_graph


def _make_one_value_graph(val: IntermediaryValue, res_node: BNode | URIRef, iri_lookups: IRILookups) -> Graph:
    match val:
        case (
            IntermediaryBoolean()
            | IntermediaryColor()
            | IntermediaryDecimal()
            | IntermediaryGeometry()
            | IntermediaryGeoname()
            | IntermediaryInt()
            | IntermediaryTime()
            | IntermediaryUri()
            | IntermediarySimpleText()
        ):
            literal_info = RDF_LITERAL_PROP_TYPE_MAPPER[type(val)]
            properties_graph = _make_value_graph_with_literal_object(
                val=val,
                res_node=res_node,
                prop_type_info=literal_info,
            )
        case IntermediaryList():
            properties_graph = _make_list_value_graph(val=val, res_node=res_node, prop_type_info=LIST_PROP_TYPE_INFO)
        case IntermediaryLink():
            target_iri = _resolve_id_to_iri(val.value, iri_lookups.id_to_iri)
            properties_graph = make_link_value_graph(
                val=val,
                val_node=BNode(),
                res_node=res_node,
                target_iri=URIRef(target_iri),
            )
        case IntermediaryRichtext():
            properties_graph = make_richtext_value_graph(
                val=val,
                val_node=BNode(),
                res_node=res_node,
                iri_resolver=iri_lookups.id_to_iri,
            )
        case IntermediaryDate():
            properties_graph = _make_date_value_graph(
                val=val,
                res_node=res_node,
            )
        case IntermediaryInterval():
            properties_graph = _make_interval_value_graph(
                val=val,
                res_node=res_node,
            )
        case _:
            raise InputError(f"Unknown value type: {type(val).__name__}")
    return properties_graph


def _make_base_value_graph(
    val: IntermediaryValue,
    val_node: BNode | URIRef,
    prop_type_info: RDFPropTypeInfo,
    res_node: BNode | URIRef,
) -> Graph:
    g = _add_optional_triples(val_node, val.permissions, val.comment)
    g.add((res_node, URIRef(val.prop_iri), val_node))
    g.add((val_node, RDF.type, prop_type_info.knora_type))
    return g


def _add_optional_triples(val_bn: BNode | URIRef, permissions: Permissions | None, comment: str | None) -> Graph:
    g = Graph()
    if permissions:
        g.add((val_bn, KNORA_API.hasPermissions, Literal(str(permissions), datatype=XSD.string)))
    if comment:
        g.add((val_bn, KNORA_API.valueHasComment, Literal(comment, datatype=XSD.string)))
    return g


def _make_value_graph_with_literal_object(
    val: LiteralValueTypesAlias,
    prop_type_info: RDFPropTypeInfo,
    res_node: BNode | URIRef,
) -> Graph:
    val_bn = BNode()
    g = _make_base_value_graph(val=val, val_node=val_bn, prop_type_info=prop_type_info, res_node=res_node)
    g.add((val_bn, prop_type_info.knora_prop, Literal(val.value, datatype=prop_type_info.xsd_type)))
    return g


def _make_list_value_graph(
    val: IntermediaryList,
    res_node: BNode | URIRef,
    prop_type_info: RDFPropTypeInfo,
) -> Graph:
    val_bn = BNode()
    g = _make_base_value_graph(val=val, val_node=val_bn, prop_type_info=prop_type_info, res_node=res_node)
    g.add((val_bn, prop_type_info.knora_prop, URIRef(val.value)))
    return g


def make_link_value_graph(
    val: IntermediaryLink,
    val_node: BNode | URIRef,
    res_node: BNode | URIRef,
    target_iri: URIRef,
) -> Graph:
    """Make a LinkValue Graph"""
    g = _make_base_value_graph(val=val, val_node=val_node, prop_type_info=LINK_PROP_TYPE_INFO, res_node=res_node)
    g.add((val_node, LINK_PROP_TYPE_INFO.knora_prop, target_iri))
    return g


def _resolve_id_to_iri(value: str, iri_resolver: IriResolver) -> URIRef:
    if is_resource_iri(value):
        return URIRef(value)
    elif resolved_iri := iri_resolver.get(value):
        return URIRef(resolved_iri)
    msg = (
        f"Could not find the ID {value} in the id2iri mapping. "
        f"This is probably because the resource '{value}' could not be created. "
        f"See {WARNINGS_SAVEPATH} for more information."
    )
    raise BaseError(msg)


def _make_date_value_graph(
    val: IntermediaryDate,
    res_node: BNode | URIRef,
) -> Graph:
    val_bn = BNode()
    date = val.value
    g = _add_optional_triples(val_bn, val.permissions, val.comment)
    g.add((res_node, URIRef(val.prop_iri), val_bn))
    g.add((val_bn, RDF.type, KNORA_API.DateValue))
    if cal := date.calendar:
        g.add((val_bn, KNORA_API.dateValueHasCalendar, Literal(cal.value, datatype=XSD.string)))
    g += _make_single_date_graph(val_bn, date.start, StartEnd.START)
    if date.end:
        g += _make_single_date_graph(val_bn, date.end, StartEnd.END)
    return g


def _make_single_date_graph(val_bn: BNode, date: SingleDate, start_end: StartEnd) -> Graph:
    def get_prop(precision: DayMonthYearEra) -> URIRef:
        return KNORA_API[f"dateValueHas{start_end.value}{precision.value}"]

    g = Graph()
    if yr := date.year:
        g.add((val_bn, get_prop(DayMonthYearEra.YEAR), Literal(yr, datatype=XSD.integer)))
    if mnt := date.month:
        g.add((val_bn, get_prop(DayMonthYearEra.MONTH), Literal(mnt, datatype=XSD.integer)))
    if day := date.day:
        g.add((val_bn, get_prop(DayMonthYearEra.DAY), Literal(day, datatype=XSD.integer)))
    if era := date.era:
        g.add((val_bn, get_prop(DayMonthYearEra.ERA), Literal(era.value, datatype=XSD.string)))
    return g


def _make_interval_value_graph(
    val: IntermediaryInterval,
    res_node: BNode | URIRef,
) -> Graph:
    val_bn = BNode()
    g = _add_optional_triples(val_bn, val.permissions, val.comment)
    g.add((res_node, URIRef(val.prop_iri), val_bn))
    g.add((val_bn, RDF.type, KNORA_API.IntervalValue))
    g.add((val_bn, KNORA_API.intervalValueHasStart, Literal(val.value.start, datatype=XSD.decimal)))
    g.add((val_bn, KNORA_API.intervalValueHasEnd, Literal(val.value.end, datatype=XSD.decimal)))
    return g


def make_richtext_value_graph(
    val: IntermediaryRichtext,
    val_node: BNode | URIRef,
    res_node: BNode | URIRef,
    iri_resolver: IriResolver,
) -> Graph:
    """
    Creates an rdflib graph for a richtext value.

    Args:
        val: Richtext value
        val_node: IRI or blank node of the value
        res_node: IRI or blank node of the resource
        iri_resolver: id to IRI resolver

    Returns:
        Graph
    """
    g = _make_base_value_graph(val=val, val_node=val_node, prop_type_info=RICHTEXT_PROP_TYPE_INFO, res_node=res_node)
    xml_with_iris = val.value.with_iris(iri_resolver)
    val_str = xml_with_iris.as_xml()
    g.add((val_node, RICHTEXT_PROP_TYPE_INFO.knora_prop, Literal(val_str, datatype=XSD.string)))
    g.add((val_node, KNORA_API.textValueHasMapping, URIRef("http://rdfh.ch/standoff/mappings/StandardMapping")))
    return g



import pandas as pd
from pandas.testing import assert_frame_equal

from dsp_tools.commands.xmlupload.models.lookup_models import TextValueData
from dsp_tools.commands.xmlupload.models.ontology_problem_models import InvalidOntologyElementsInData
from dsp_tools.commands.xmlupload.models.ontology_problem_models import InvalidTextValueEncodings
from dsp_tools.commands.xmlupload.models.ontology_problem_models import _make_msg_for_one_resource
from dsp_tools.commands.xmlupload.models.ontology_problem_models import _make_msg_from_df


class TestInvalidOntologyElementsInData:
    def test_print_problem_string_cls(self) -> None:
        onto = InvalidOntologyElementsInData([("clsA", ["idA"], "wrong")], [], ontos_on_server=["test"])
        msg = onto._compose_problem_string_for_cls()
        assert msg == (
            "The following resource(s) have an invalid resource type:\n\n"
            "    Resource Type: 'clsA'\n"
            "    Problem: 'wrong'\n"
            "    Resource ID(s):\n"
            "    - idA\n"
        )

    def test_print_problem_string_no_cls(self) -> None:
        onto = InvalidOntologyElementsInData([], [], [])
        assert not onto._compose_problem_string_for_cls()

    def test_print_problem_string_prop(self) -> None:
        onto = InvalidOntologyElementsInData([], [("propA", ["idA"], "wrong")], ["test"])
        msg = onto._compose_problem_string_for_props()
        assert msg == (
            "The following resource(s) have invalid property type(s):\n\n"
            "    Property Name: 'propA'\n"
            "    Problem: 'wrong'\n"
            "    Resource ID(s):\n"
            "    - idA\n"
        )

    def test_print_problem_string_no_prop(self) -> None:
        onto = InvalidOntologyElementsInData([], [], [])
        assert not onto._compose_problem_string_for_props()

    def test_execute_problem_protocol(self) -> None:
        onto = InvalidOntologyElementsInData(
            [("clsA", ["idA"], "wrong")],
            [("propA", ["idA"], "wrong"), ("propB", ["idB", "idC"], "wrong")],
            ["test1", "test2"],
        )
        expected_msg = (
            "\nSome property and/or class type(s) used in the XML are unknown.\n"
            "The ontologies for your project on the server are:\n"
            "    - test1\n"
            "    - test2"
            "\n\n"
            "The following resource(s) have an invalid resource type:\n\n"
            "    Resource Type: 'clsA'\n"
            "    Problem: 'wrong'\n"
            "    Resource ID(s):\n"
            "    - idA"
            "\n\n"
            "The following resource(s) have invalid property type(s):\n\n"
            "    Property Name: 'propA'\n"
            "    Problem: 'wrong'\n"
            "    Resource ID(s):\n"
            "    - idA"
            "\n\n"
            "    Property Name: 'propB'\n"
            "    Problem: 'wrong'\n"
            "    Resource ID(s):\n"
            "    - idB\n"
            "    - idC"
            "\n\n"
        )
        msg, df = onto.execute_problem_protocol()
        assert not df
        assert msg == expected_msg

    def test_get_problems_as_df(self) -> None:
        onto = InvalidOntologyElementsInData(
            [("clsA", ["idA"], "wrongA")],
            [("propB", ["idB"], "wrongB"), ("propC", ["idC1", "idC2"], "wrongC")],
            ["test"],
        )
        expected_df = pd.DataFrame(
            {
                "problematic type": ["clsA", "propB", "propC", "propC"],
                "resource id": ["idA", "idB", "idC1", "idC2"],
                "problem": ["wrongA", "wrongB", "wrongC", "wrongC"],
            }
        )
        res_df = onto._get_problems_as_df()
        assert_frame_equal(res_df, expected_df)


class TestInvalidTextValueEncodings:
    def test_get_problems_as_df(self) -> None:
        problems = InvalidTextValueEncodings(
            [
                TextValueData("id1", ":restype", ":simple", "xml"),
                TextValueData("id1", ":restype", ":rich", "utf8"),
                TextValueData("id2", ":restype", ":rich", "utf8"),
            ]
        )
        expected_df = pd.DataFrame(
            {
                "Resource ID": ["id1", "id1", "id2"],
                "Resource Type": [":restype", ":restype", ":restype"],
                "Property Name": [":rich", ":simple", ":rich"],
                "Encoding Used": ["utf8", "xml", "utf8"],
            }
        )
        res_df = problems._get_problems_as_df()
        assert_frame_equal(res_df, expected_df)

    def test_make_msg_for_one_resource(self) -> None:
        test_df = pd.DataFrame(
            {
                "Resource ID": ["id1", "id1"],
                "Resource Type": [":restype", ":restype"],
                "Property Name": [":rich", ":simple"],
                "Encoding Used": ["utf8", "xml"],
            }
        )
        res = _make_msg_for_one_resource("id1", test_df)
        expected = (
            "Resource ID: 'id1' | Resource Type: ':restype'\n"
            "    - Property Name: ':rich' -> Encoding Used: 'utf8'\n"
            "    - Property Name: ':simple' -> Encoding Used: 'xml'"
        )
        assert res == expected

    def test_make_msg_from_df(self) -> None:
        test_df = pd.DataFrame(
            {
                "Resource ID": ["id1", "id1", "id2"],
                "Resource Type": [":restype", ":restype", ":restype"],
                "Property Name": [":rich", ":simple", ":rich"],
                "Encoding Used": ["utf8", "xml", "utf8"],
            }
        )
        res = _make_msg_from_df(test_df)
        expected = (
            "Resource ID: 'id1' | Resource Type: ':restype'\n"
            "    - Property Name: ':rich' -> Encoding Used: 'utf8'\n"
            "    - Property Name: ':simple' -> Encoding Used: 'xml'"
            "\n\n"
            "Resource ID: 'id2' | Resource Type: ':restype'\n"
            "    - Property Name: ':rich' -> Encoding Used: 'utf8'"
        )
        assert res == expected



from loguru import logger
from rdflib import Graph

from dsp_tools.commands.validate_data.models.api_responses import AllProjectLists
from dsp_tools.commands.validate_data.models.validation import SHACLGraphs
from dsp_tools.commands.validate_data.sparql.cardinality_shacl import construct_cardinality_node_shapes
from dsp_tools.commands.validate_data.sparql.legal_info_shacl import construct_allowed_licenses_shape
from dsp_tools.commands.validate_data.sparql.value_shacl import construct_property_shapes


def construct_shapes_graphs(onto: Graph, knora_api: Graph, project_lists: AllProjectLists) -> SHACLGraphs:
    """
    Constructs a shapes graph from a project ontology

    Args:
        onto: ontology as graph
        knora_api: the knora-api ontology
        project_lists: the lists of a project

    Returns:
        shapes graph
    """
    logger.info("Constructing SHACL shapes from ontology.")
    knora_subset = _get_all_relevant_knora_subset(knora_api)
    graph_to_query = knora_subset + onto
    cardinality = construct_cardinality_node_shapes(graph_to_query)
    content = construct_property_shapes(graph_to_query, project_lists)
    content += construct_allowed_licenses_shape()
    return SHACLGraphs(cardinality=cardinality, content=content)


def _get_all_relevant_knora_subset(knora_api: Graph) -> Graph:
    logger.info("Getting relevant knora-api subset")
    props = ["isLinkProperty", "isEditable", "isLinkValueProperty"]
    g = Graph()
    for p in props:
        g += _get_one_relevant_knora_subset(knora_api, p)
    return g


def _get_one_relevant_knora_subset(knora_api: Graph, knora_prop: str) -> Graph:
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>

    CONSTRUCT {
    
      ?focusProp a owl:ObjectProperty ;
                 knora-api:objectType ?type ;
                 knora-api:%(knora_prop)s ?object .
      
    } WHERE {
      
      ?focusProp a owl:ObjectProperty ;
                 knora-api:objectType ?type ;
                 knora-api:%(knora_prop)s ?object .
      
    }
    """ % {"knora_prop": knora_prop}  # noqa: UP031 (printf-string-formatting)
    if results_graph := knora_api.query(query_s).graph:
        return results_graph
    return Graph()



from __future__ import annotations

import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Any
from typing import Protocol

from dsp_tools.error.custom_warnings import DspToolsUserWarning
from dsp_tools.utils.data_formats.uri_util import is_iiif_uri
from dsp_tools.xmllib.models.config_options import Permissions
from dsp_tools.xmllib.value_checkers import is_string_like


@dataclass
class AuthorshipLookup:
    lookup: dict[tuple[str, ...], str]

    def get_id(self, authors: tuple[str, ...]) -> str:
        if not (found := self.lookup.get(authors)):
            warnings.warn(DspToolsUserWarning(f"The input authors {authors} are not defined in the look-up."))
            return " / ".join([str(x) for x in authors])
        return found


@dataclass
class Metadata:
    license: str
    copyright_holder: str
    authorship: tuple[str, ...]
    permissions: Permissions
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_string_like(self.license):
            _warn_type_mismatch(expected_type="license", value=self.license, res_id=self.resource_id)
        if not is_string_like(str(self.copyright_holder)):
            _warn_type_mismatch(expected_type="copyright holder", value=self.copyright_holder, res_id=self.resource_id)
        if len(self.authorship) == 0:
            _warn_type_mismatch(
                expected_type="list of authorship strings", value="empty input", res_id=self.resource_id
            )
        for author in self.authorship:
            if not is_string_like(author):
                _warn_type_mismatch(expected_type="author", value=author, res_id=self.resource_id)


class AbstractFileValue(Protocol):
    value: str | Path
    metadata: Metadata
    comment: str | None


@dataclass
class FileValue(AbstractFileValue):
    value: str | Path
    metadata: Metadata
    comment: str | None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_string_like(str(self.value)):
            _warn_type_mismatch(expected_type="file name", value=self.value, res_id=self.resource_id)


@dataclass
class IIIFUri(AbstractFileValue):
    value: str
    metadata: Metadata
    comment: str | None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_iiif_uri(self.value):
            _warn_type_mismatch(expected_type="IIIF uri", value=self.value, res_id=self.resource_id)


def _warn_type_mismatch(expected_type: str, value: Any, res_id: str | None) -> None:
    """Emits a warning if a values is not in the expected format."""
    if res_id:
        msg = f"The Resource '{res_id}' has an invalid input: The value '{value}' is not a valid {expected_type}."
    else:
        msg = f"The value '{value}' is not a valid {expected_type}."
    warnings.warn(DspToolsUserWarning(msg))



import pytest
from rdflib import Graph

from test.unittests.commands.validate_data.constants import PREFIXES


@pytest.fixture
def onto_for_cardinality() -> Graph:
    g = Graph()
    g.parse("testdata/validate-data/onto_for_cardinality_checks.ttl")
    return g


@pytest.fixture
def one_res_one_prop() -> Graph:
    ttl = f"""{PREFIXES}
    onto:CardOneResource a owl:Class ;
    rdfs:label "Resource with One Cardinality" ;
    knora-api:canBeInstantiated true ;
    knora-api:isResourceClass true ;
    rdfs:subClassOf 
        knora-api:Resource ,
        [ a owl:Restriction ;
            knora-api:isInherited true ;
            owl:maxCardinality 1 ;
            owl:onProperty knora-api:versionDate ],
        [ a owl:Restriction ;
            knora-api:isInherited true ;
            salsah-gui:guiOrder 0 ;
            owl:cardinality 1 ;
            owl:onProperty onto:testBoolean ] .

    onto:testBoolean a owl:ObjectProperty ;
        rdfs:label "Test Boolean" ;
        knora-api:isEditable true ;
        knora-api:isResourceProperty true ;
        knora-api:objectType knora-api:BooleanValue ;
        salsah-gui:guiElement salsah-gui:Checkbox ;
        rdfs:subPropertyOf knora-api:hasValue .
    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g


@pytest.fixture
def one_prop() -> Graph:
    ttl = f"""{PREFIXES}
    onto:testBoolean a owl:ObjectProperty ;
        rdfs:label "Test Boolean" ;
        knora-api:isEditable true ;
        knora-api:isResourceProperty true ;
        knora-api:objectType knora-api:BooleanValue ;
        salsah-gui:guiElement salsah-gui:Checkbox ;
        rdfs:subPropertyOf knora-api:hasValue .
    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g


@pytest.fixture
def link_prop() -> Graph:
    ttl = f"""{PREFIXES}
    onto:testHasLinkTo a owl:ObjectProperty ;
        rdfs:label "Test In-built hasLinkTo" ;
        knora-api:isEditable true ;
        knora-api:isLinkProperty true ;
        knora-api:isResourceProperty true ;
        knora-api:objectType knora-api:Resource ;
        salsah-gui:guiElement salsah-gui:Searchbox ;
        rdfs:subPropertyOf knora-api:hasLinkTo .
    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g


@pytest.fixture
def card_1() -> Graph:
    ttl = f"""{PREFIXES}
    onto:ClassMixedCard a owl:Class ;
        knora-api:isResourceClass true ;
        knora-api:canBeInstantiated true ;
        rdfs:subClassOf [ 
                a owl:Restriction ;
                salsah-gui:guiOrder 0 ;
                owl:cardinality 1 ;
                owl:onProperty onto:testBoolean
                         ] .
    onto:testBoolean a owl:ObjectProperty ;
    rdfs:label "Test Boolean" ;
    knora-api:isEditable true ;
    knora-api:isResourceProperty true ;
    knora-api:objectType knora-api:BooleanValue ;
    salsah-gui::guiElement salsah-gui::Checkbox ;
    rdfs:subPropertyOf knora-api:hasValue .
    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g


@pytest.fixture
def card_0_1() -> Graph:
    ttl = f"""{PREFIXES}
    onto:ClassMixedCard a owl:Class ;
        knora-api:isResourceClass true ;
        knora-api:canBeInstantiated true ;
        rdfs:subClassOf [ 
                a owl:Restriction ;
                salsah-gui:guiOrder 1 ;
                owl:maxCardinality 1 ;
                owl:onProperty onto:testDecimalSimpleText
                         ] . 
    onto:testDecimalSimpleText knora-api:isEditable true .
    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g


@pytest.fixture
def card_1_n() -> Graph:
    ttl = f"""{PREFIXES}
    onto:ClassMixedCard a owl:Class ;
        knora-api:isResourceClass true ;
        knora-api:canBeInstantiated true ;
        rdfs:subClassOf [ 
                a owl:Restriction ;
                salsah-gui:guiOrder 2 ;
                owl:minCardinality 1 ;
                owl:onProperty onto:testGeoname
                         ] .
    onto:testGeoname knora-api:isEditable true .
    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g


@pytest.fixture
def card_0_n() -> Graph:
    ttl = f"""{PREFIXES}
    onto:ClassMixedCard a owl:Class ;
        knora-api:isResourceClass true ;
        knora-api:canBeInstantiated true ;
        rdfs:subClassOf [ 
                a owl:Restriction ;
                salsah-gui:guiOrder 3 ;
                owl:minCardinality 0 ;
                owl:onProperty onto:testSimpleText
                         ] .
    onto:testSimpleText knora-api:isEditable true ;
                        knora-api:objectType knora-api:TextValue .
    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g


@pytest.fixture
def one_bool_prop() -> Graph:
    ttl = f"""{PREFIXES}
    onto:testBoolean a owl:ObjectProperty ;
        rdfs:label "Test Boolean" ;
        knora-api:isEditable true ;
        knora-api:isResourceProperty true ;
        knora-api:objectType knora-api:BooleanValue ;
        salsah-gui:guiElement salsah-gui:Checkbox ;
        rdfs:subPropertyOf knora-api:hasValue .
    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g


@pytest.fixture
def one_richtext_prop() -> Graph:
    ttl = f"""{PREFIXES}
    onto:testRichtext a owl:ObjectProperty ;
        rdfs:label "Test Richtext" ;
        knora-api:isEditable true ;
        knora-api:isResourceProperty true ;
        knora-api:objectType knora-api:TextValue ;
        salsah-gui:guiElement salsah-gui:Richtext ;
        rdfs:subPropertyOf knora-api:hasValue .
    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g


@pytest.fixture
def link_prop_card_1() -> Graph:
    ttl = f"""{PREFIXES}
    onto:ClassMixedCard a owl:Class ;
    rdfs:label "Resource with all cardinality options" ;
    knora-api:canBeInstantiated true ;
    knora-api:isResourceClass true ;
    rdfs:subClassOf [ 
            a owl:Restriction ;
            salsah-gui:guiOrder 1 ;
            owl:cardinality 1 ;
            owl:onProperty onto:testHasLinkToCardOneResourceValue 
            ] ,
            [ a owl:Restriction ;
            salsah-gui:guiOrder 1 ;
            owl:cardinality 1 ;
            owl:onProperty onto:testHasLinkToCardOneResource ] .
            
    onto:testHasLinkToCardOneResource a owl:ObjectProperty ;
        rdfs:label "Super-class" ;
        knora-api:isEditable true ;
        knora-api:isLinkProperty true ;
        knora-api:isResourceProperty true ;
        knora-api:objectType onto:CardOneResource ;
        salsah-gui:guiElement salsah-gui:Searchbox ;
        rdfs:subPropertyOf knora-api:hasLinkTo .
    
    onto:testHasLinkToCardOneResourceValue a owl:ObjectProperty ;
        rdfs:label "Super-class" ;
        knora-api:isEditable true ;
        knora-api:isLinkValueProperty true ;
        knora-api:isResourceProperty true ;
        knora-api:objectType knora-api:LinkValue ;
        salsah-gui:guiElement salsah-gui:Searchbox ;
        rdfs:subPropertyOf knora-api:hasLinkToValue .

    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g


@pytest.fixture
def link_prop_card_01() -> Graph:
    ttl = f"""{PREFIXES}
    onto:ClassMixedCard a owl:Class ;
    rdfs:label "Resource with all cardinality options" ;
    knora-api:canBeInstantiated true ;
    knora-api:isResourceClass true ;
    rdfs:subClassOf [ 
            a owl:Restriction ;
            salsah-gui:guiOrder 1 ;
            owl:maxCardinality 1 ;
            owl:onProperty onto:testHasLinkToCardOneResourceValue 
            ] ,
            [ a owl:Restriction ;
            salsah-gui:guiOrder 1 ;
            owl:maxCardinality 1 ;
            owl:onProperty onto:testHasLinkToCardOneResource ] .

    onto:testHasLinkToCardOneResource a owl:ObjectProperty ;
        rdfs:label "Super-class" ;
        knora-api:isEditable true ;
        knora-api:isLinkProperty true ;
        knora-api:isResourceProperty true ;
        knora-api:objectType onto:CardOneResource ;
        salsah-gui:guiElement salsah-gui:Searchbox ;
        rdfs:subPropertyOf knora-api:hasLinkTo .

    onto:testHasLinkToCardOneResourceValue a owl:ObjectProperty ;
        rdfs:label "Super-class" ;
        knora-api:isEditable true ;
        knora-api:isLinkValueProperty true ;
        knora-api:isResourceProperty true ;
        knora-api:objectType knora-api:LinkValue ;
        salsah-gui:guiElement salsah-gui:Searchbox ;
        rdfs:subPropertyOf knora-api:hasLinkToValue .

    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g



from uuid import uuid4

from dsp_tools.commands.xmlupload.models.intermediary.file_values import IntermediaryFileMetadata
from dsp_tools.commands.xmlupload.models.intermediary.file_values import IntermediaryFileValue
from dsp_tools.commands.xmlupload.models.intermediary.file_values import IntermediaryIIIFUri
from dsp_tools.commands.xmlupload.models.intermediary.res import IntermediaryResource
from dsp_tools.commands.xmlupload.models.intermediary.res import MigrationMetadata
from dsp_tools.commands.xmlupload.models.intermediary.res import ResourceInputConversionFailure
from dsp_tools.commands.xmlupload.models.intermediary.res import ResourceTransformationResult
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryBoolean
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryColor
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryDate
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryDecimal
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryGeometry
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryGeoname
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryInt
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryInterval
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryLink
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryList
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryRichtext
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediarySimpleText
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryTime
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryUri
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryValue
from dsp_tools.commands.xmlupload.models.lookup_models import IntermediaryLookups
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.prepare_xml_input.ark2iri import convert_ark_v0_to_resource_iri
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import TypeTransformerMapper
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import assert_is_string
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import assert_is_tuple
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_boolean
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_date
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_decimal
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_geometry
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_integer
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_interval
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_richtext
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_simpletext
from dsp_tools.error.exceptions import InputError
from dsp_tools.error.exceptions import PermissionNotExistsError
from dsp_tools.legacy_models.datetimestamp import DateTimeStamp
from dsp_tools.utils.xml_parsing.models.parsed_resource import KnoraValueType
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedFileValue
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedFileValueMetadata
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedMigrationMetadata
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedResource
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedValue

TYPE_TRANSFORMER_MAPPER: dict[KnoraValueType, TypeTransformerMapper] = {
    KnoraValueType.BOOLEAN_VALUE: TypeTransformerMapper(IntermediaryBoolean, transform_boolean),
    KnoraValueType.COLOR_VALUE: TypeTransformerMapper(IntermediaryColor, assert_is_string),
    KnoraValueType.DECIMAL_VALUE: TypeTransformerMapper(IntermediaryDecimal, transform_decimal),
    KnoraValueType.DATE_VALUE: TypeTransformerMapper(IntermediaryDate, transform_date),
    KnoraValueType.GEOM_VALUE: TypeTransformerMapper(IntermediaryGeometry, transform_geometry),
    KnoraValueType.GEONAME_VALUE: TypeTransformerMapper(IntermediaryGeoname, assert_is_string),
    KnoraValueType.INT_VALUE: TypeTransformerMapper(IntermediaryInt, transform_integer),
    KnoraValueType.INTERVAL_VALUE: TypeTransformerMapper(IntermediaryInterval, transform_interval),
    KnoraValueType.TIME_VALUE: TypeTransformerMapper(IntermediaryTime, assert_is_string),
    KnoraValueType.SIMPLETEXT_VALUE: TypeTransformerMapper(IntermediarySimpleText, transform_simpletext),
    KnoraValueType.URI_VALUE: TypeTransformerMapper(IntermediaryUri, assert_is_string),
}


def transform_all_resources_into_intermediary_resources(
    resources: list[ParsedResource], lookups: IntermediaryLookups
) -> ResourceTransformationResult:
    failures = []
    transformed = []
    for res in resources:
        try:
            result = _transform_one_resource(res, lookups)
            transformed.append(result)
        except (PermissionNotExistsError, InputError) as e:
            failures.append(ResourceInputConversionFailure(res.res_id, str(e)))
    return ResourceTransformationResult(transformed, failures)


def _transform_one_resource(resource: ParsedResource, lookups: IntermediaryLookups) -> IntermediaryResource:
    permissions = _resolve_permission(resource.permissions_id, lookups.permissions)
    values = [_transform_one_value(val, lookups) for val in resource.values]
    file_val, iiif_uri, migration_metadata = None, None, None
    if resource.file_value:
        if resource.file_value.value_type == KnoraValueType.STILL_IMAGE_IIIF:
            iiif_uri = _transform_iiif_uri_value(resource.file_value, lookups)
        else:
            file_val = _transform_file_value(
                val=resource.file_value, lookups=lookups, res_id=resource.res_id, res_label=resource.label
            )
    if resource.migration_metadata:
        migration_metadata = _transform_migration_metadata(resource.migration_metadata)
    return IntermediaryResource(
        res_id=resource.res_id,
        type_iri=resource.res_type,
        label=resource.label,
        permissions=permissions,
        values=values,
        file_value=file_val,
        iiif_uri=iiif_uri,
        migration_metadata=migration_metadata,
    )


def _transform_migration_metadata(metadata: ParsedMigrationMetadata) -> MigrationMetadata:
    res_iri = metadata.iri
    # ARK takes precedence over the IRI,
    # but must be transformed into an IRI as it is only for external reference and not consistent with a DB IRI
    if metadata.ark:
        res_iri = convert_ark_v0_to_resource_iri(metadata.ark)
    date = DateTimeStamp(metadata.creation_date) if metadata.creation_date else None
    return MigrationMetadata(res_iri, date)


def _transform_file_value(
    val: ParsedFileValue, lookups: IntermediaryLookups, res_id: str, res_label: str
) -> IntermediaryFileValue:
    metadata = _get_metadata(val.metadata, lookups)
    file_val = assert_is_string(val.value)
    return IntermediaryFileValue(value=file_val, metadata=metadata, res_id=res_id, res_label=res_label)


def _transform_iiif_uri_value(iiif_uri: ParsedFileValue, lookups: IntermediaryLookups) -> IntermediaryIIIFUri:
    metadata = _get_metadata(iiif_uri.metadata, lookups)
    file_val = assert_is_string(iiif_uri.value)
    return IntermediaryIIIFUri(file_val, metadata)


def _get_metadata(file_metadata: ParsedFileValueMetadata, lookups: IntermediaryLookups) -> IntermediaryFileMetadata:
    permissions = _resolve_permission(file_metadata.permissions_id, lookups.permissions)
    predefined_licenses = [
        "http://rdfh.ch/licenses/cc-by-4.0",
        "http://rdfh.ch/licenses/cc-by-sa-4.0",
        "http://rdfh.ch/licenses/cc-by-nc-4.0",
        "http://rdfh.ch/licenses/cc-by-nc-sa-4.0",
        "http://rdfh.ch/licenses/cc-by-nd-4.0",
        "http://rdfh.ch/licenses/cc-by-nc-nd-4.0",
        "http://rdfh.ch/licenses/ai-generated",
        "http://rdfh.ch/licenses/unknown",
        "http://rdfh.ch/licenses/public-domain",
    ]
    if file_metadata.license_iri and file_metadata.license_iri not in predefined_licenses:
        raise InputError(
            f"The license '{file_metadata.license_iri}' used for an image or iiif-uri is unknown. "
            f"See documentation for accepted pre-defined licenses."
        )
    return IntermediaryFileMetadata(
        license_iri=file_metadata.license_iri,
        copyright_holder=file_metadata.copyright_holder,
        authorships=_resolve_authorship(file_metadata.authorship_id, lookups.authorships),
        permissions=permissions,
    )


def _resolve_authorship(authorship_id: str | None, lookup: dict[str, list[str]]) -> list[str] | None:
    if not authorship_id:
        return None
    if not (found := lookup.get(authorship_id)):
        raise InputError(
            f"The authorship id '{authorship_id}' referenced in a multimedia file or iiif-uri is unknown. "
            f"Ensure that all referenced ids are defined in the `<authorship>` elements of your XML file."
        )
    return found


def _transform_one_value(val: ParsedValue, lookups: IntermediaryLookups) -> IntermediaryValue:
    match val.value_type:
        case KnoraValueType.LIST_VALUE:
            return _transform_list_value(val, lookups)
        case KnoraValueType.LINK_VALUE:
            return _transform_link_value(val, lookups)
        case KnoraValueType.RICHTEXT_VALUE:
            return _transform_richtext_value(val, lookups)
        case _ as val_type:
            transformation_mapper = TYPE_TRANSFORMER_MAPPER[val_type]
            return _transform_generic_value(val=val, lookups=lookups, transformation_mapper=transformation_mapper)


def _transform_generic_value(
    val: ParsedValue, lookups: IntermediaryLookups, transformation_mapper: TypeTransformerMapper
) -> IntermediaryValue:
    transformed_value = transformation_mapper.val_transformer(val.value)
    permission_val = _resolve_permission(val.permissions_id, lookups.permissions)
    return transformation_mapper.val_type(transformed_value, val.prop_name, val.comment, permission_val)


def _transform_link_value(val: ParsedValue, lookups: IntermediaryLookups) -> IntermediaryValue:
    transformed_value = assert_is_string(val.value)
    permission_val = _resolve_permission(val.permissions_id, lookups.permissions)
    link_val: IntermediaryValue = IntermediaryLink(
        value=transformed_value,
        prop_iri=val.prop_name,
        comment=val.comment,
        permissions=permission_val,
        value_uuid=str(uuid4()),
    )
    return link_val


def _transform_list_value(val: ParsedValue, lookups: IntermediaryLookups) -> IntermediaryValue:
    tuple_val = assert_is_tuple(val.value)
    if not (list_iri := lookups.listnodes.get(tuple_val)):
        raise InputError(f"Could not find list iri for node: {' / '.join(tuple_val)}")
    permission_val = _resolve_permission(val.permissions_id, lookups.permissions)
    list_val: IntermediaryValue = IntermediaryList(
        value=list_iri,
        prop_iri=val.prop_name,
        comment=val.comment,
        permissions=permission_val,
    )
    return list_val


def _transform_richtext_value(val: ParsedValue, lookups: IntermediaryLookups) -> IntermediaryValue:
    transformed_value = transform_richtext(val.value)
    permission_val = _resolve_permission(val.permissions_id, lookups.permissions)
    richtext: IntermediaryValue = IntermediaryRichtext(
        value=transformed_value,
        prop_iri=val.prop_name,
        comment=val.comment,
        permissions=permission_val,
        resource_references=transformed_value.find_internal_ids(),
        value_uuid=str(uuid4()),
    )
    return richtext


def _resolve_permission(permissions: str | None, permissions_lookup: dict[str, Permissions]) -> Permissions | None:
    """Resolve the permission into a string that can be sent to the API."""
    if permissions:
        if not (per := permissions_lookup.get(permissions)):
            raise PermissionNotExistsError(f"Could not find permissions for value: {permissions}")
        return per
    return None



from __future__ import annotations

from dataclasses import dataclass

from dsp_tools.commands.xmlupload.models.permission import Permissions


@dataclass
class IntermediaryFileMetadata:
    license_iri: str | None
    copyright_holder: str | None
    authorships: list[str] | None
    permissions: Permissions | None = None

    def all_legal_info(self) -> bool:
        return all([self.license_iri, self.copyright_holder, self.authorships])


@dataclass
class IntermediaryFileValue:
    value: str
    metadata: IntermediaryFileMetadata
    res_id: str
    res_label: str


@dataclass
class IntermediaryIIIFUri:
    value: str
    metadata: IntermediaryFileMetadata



from __future__ import annotations

import subprocess
from contextlib import contextmanager
from pathlib import Path
from typing import Iterator
from uuid import uuid4

import regex
from testcontainers.core.network import Network

from test.e2e.setup_testcontainers.artifacts import get_artifact_dirs
from test.e2e.setup_testcontainers.artifacts import remove_artifact_dirs
from test.e2e.setup_testcontainers.containers import ContainerMetadata
from test.e2e.setup_testcontainers.containers import ContainerNames
from test.e2e.setup_testcontainers.containers import Containers
from test.e2e.setup_testcontainers.containers import ImageVersions
from test.e2e.setup_testcontainers.containers import get_all_containers
from test.e2e.setup_testcontainers.ports import get_ports
from test.e2e.setup_testcontainers.ports import release_ports


@contextmanager
def get_containers() -> Iterator[ContainerMetadata]:
    if subprocess.run("docker stats --no-stream".split(), check=False).returncode != 0:
        raise RuntimeError("Docker is not running properly")
    with Network() as network:
        metadata = _get_container_metadata()
        containers = get_all_containers(network, metadata)
        try:
            yield metadata
        finally:
            _stop_all_containers(containers)
            release_ports(metadata.ports)
            remove_artifact_dirs(metadata.artifact_dirs)


def _get_container_metadata() -> ContainerMetadata:
    _uuid = str(uuid4())[:6]
    return ContainerMetadata(
        artifact_dirs=get_artifact_dirs(_uuid),
        versions=_get_image_versions(),
        ports=get_ports(),
        names=_get_container_names(_uuid),
    )


def _get_image_versions() -> ImageVersions:
    def _get_version(docker_compose_content: str, component: str) -> str:
        match = regex.search(rf"image: daschswiss/{component}:([^\n]+)", docker_compose_content)
        return match.group(1) if match else "latest"

    docker_compose_content = Path("src/dsp_tools/resources/start-stack/docker-compose.yml").read_text(encoding="utf-8")
    fuseki = _get_version(docker_compose_content, "apache-jena-fuseki")
    sipi = _get_version(docker_compose_content, "knora-sipi")
    ingest = _get_version(docker_compose_content, "dsp-ingest")
    api = _get_version(docker_compose_content, "knora-api")
    return ImageVersions(fuseki=fuseki, sipi=sipi, ingest=ingest, api=api)


def _get_container_names(_uuid: str) -> ContainerNames:
    prefix = f"testcontainer-{_uuid}"
    return ContainerNames(f"{prefix}-DB", f"{prefix}-SIPI", f"{prefix}-INGEST", f"{prefix}-API")


def _stop_all_containers(containers: Containers) -> None:
    containers.fuseki.stop()
    containers.sipi.stop()
    containers.ingest.stop()
    containers.api.stop()
    print("All containers have been stopped")



from pathlib import Path

import pytest

from dsp_tools.commands.xmlupload.models.lookup_models import IntermediaryLookups
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.prepare_xml_input.prepare_xml_input import _extract_resources_from_xml
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_xmlresource_into_intermediary_classes import (
    transform_xmlresources_into_intermediary_resources,
)
from dsp_tools.commands.xmlupload.stash.analyse_circular_reference_graph import generate_upload_order
from dsp_tools.commands.xmlupload.stash.create_info_for_graph import create_info_for_graph_from_intermediary_resources
from dsp_tools.commands.xmlupload.stash.stash_circular_references import stash_circular_references
from dsp_tools.utils.ansi_colors import RESET_TO_DEFAULT
from dsp_tools.utils.ansi_colors import YELLOW
from dsp_tools.utils.xml_parsing.parse_and_transform import parse_and_clean_xml_file


def test_get_length_ok_resources() -> None:
    test_root = parse_and_clean_xml_file(Path("testdata/xml-data/test-circular-references.xml"))
    resources = _extract_resources_from_xml(test_root, "simcir")
    permissions_lookup = {"open": Permissions()}
    intermediary_lookups = IntermediaryLookups(
        permissions_lookup, {}, namespaces={"simcir": "https://namespace.ch/simcir#"}, authorships={}
    )
    transformation_result = transform_xmlresources_into_intermediary_resources(resources, intermediary_lookups)
    info_for_graph = create_info_for_graph_from_intermediary_resources(transformation_result.transformed_resources)
    stash_lookup, _ = generate_upload_order(info_for_graph)
    stash = stash_circular_references(transformation_result.transformed_resources, stash_lookup)
    len_standoff = len(stash.standoff_stash.res_2_stash_items)  # type: ignore[union-attr]
    len_resptr = len(stash.link_value_stash.res_2_stash_items)  # type: ignore[union-attr]
    stashed_links = len_standoff + len_resptr
    previous_stash_size = 14
    print_str = (
        f"\n\n---------------------\n"
        f"Total Resources: 63\n"
        f"Previous Stash Size: {previous_stash_size}\n"
        f"Current Stash Size: {stashed_links}"
        f"\n---------------------\n"
    )
    print(YELLOW + print_str + RESET_TO_DEFAULT)
    assert stashed_links <= previous_stash_size


if __name__ == "__main__":
    pytest.main([__file__])



from pathlib import Path

from rdflib import RDF
from rdflib import XSD
from rdflib import BNode
from rdflib import Graph
from rdflib import Literal
from rdflib import URIRef

from dsp_tools.commands.xmlupload.make_rdf_graph.constants import ARCHIVE_FILE_VALUE
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import AUDIO_FILE_VALUE
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import DOCUMENT_FILE_VALUE
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import IIIF_URI_VALUE
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import MOVING_IMAGE_FILE_VALUE
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import STILL_IMAGE_FILE_VALUE
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import TEXT_FILE_VALUE
from dsp_tools.commands.xmlupload.models.deserialise.xmlresource import BitstreamInfo
from dsp_tools.commands.xmlupload.models.rdf_models import AbstractFileValue
from dsp_tools.commands.xmlupload.models.rdf_models import FileValueMetadata
from dsp_tools.commands.xmlupload.models.rdf_models import RDFPropTypeInfo
from dsp_tools.error.exceptions import BaseError
from dsp_tools.utils.rdflib_constants import KNORA_API


def make_iiif_uri_value_graph(iiif_uri: AbstractFileValue, res_node: BNode | URIRef) -> Graph:
    """
    Creates a graph with the IIIF-URI Link

    Args:
        iiif_uri: Information about the IIIF URI
        res_node: Node of the resource

    Returns:
        Graph with the IIIF-URI Value
    """
    return _make_abstract_file_value_graph(
        iiif_uri, IIIF_URI_VALUE, res_node, KNORA_API.stillImageFileValueHasExternalUrl
    )


def make_file_value_graph(
    bitstream_info: BitstreamInfo, file_value_metadata: FileValueMetadata, res_node: BNode | URIRef
) -> Graph:
    """
    Creates a graph with the File Value information.

    Args:
        bitstream_info: Information about the previously uploaded file
        file_value_metadata: Metadata of the file value
        res_node: Node of the resource

    Returns:
        Graph with the File Value
    """
    file_type = get_file_type_info(bitstream_info.local_file)
    internal_filename = bitstream_info.internal_file_name
    file_value = AbstractFileValue(internal_filename, file_value_metadata)
    return _make_abstract_file_value_graph(file_value, file_type, res_node)


def get_file_type_info(local_file: str) -> RDFPropTypeInfo:
    """
    Takes path of a file and returns the correct file value type.

    Args:
        local_file: filepath

    Returns:
        File type info to construct the graph

    Raises:
        BaseError: in case the extension is unknown
    """
    file_ending = Path(local_file).suffix[1:].lower()
    match file_ending:
        case "zip" | "tar" | "gz" | "z" | "tgz" | "gzip" | "7z":
            return ARCHIVE_FILE_VALUE
        case "mp3" | "wav":
            return AUDIO_FILE_VALUE
        case "pdf" | "doc" | "docx" | "xls" | "xlsx" | "ppt" | "pptx":
            return DOCUMENT_FILE_VALUE
        case "mp4":
            return MOVING_IMAGE_FILE_VALUE
        # jpx is the extension of the files returned by dsp-ingest
        case "jpg" | "jpeg" | "jp2" | "png" | "tif" | "tiff" | "jpx":
            return STILL_IMAGE_FILE_VALUE
        case "odd" | "rng" | "txt" | "xml" | "xsd" | "xsl" | "csv" | "json":
            return TEXT_FILE_VALUE
        case _:
            raise BaseError(f"Unknown file ending '{file_ending}' for file '{local_file}'")


def _make_abstract_file_value_graph(
    file_value: AbstractFileValue,
    type_info: RDFPropTypeInfo,
    res_node: BNode | URIRef,
    has_file_name_prop: URIRef = KNORA_API.fileValueHasFilename,
) -> Graph:
    file_bn = BNode()
    g = _add_metadata(file_bn, file_value.metadata)
    g.add((res_node, type_info.knora_prop, file_bn))
    g.add((file_bn, RDF.type, type_info.knora_type))
    g.add((file_bn, has_file_name_prop, Literal(file_value.value, datatype=XSD.string)))
    return g


def _add_metadata(file_bn: BNode, metadata: FileValueMetadata) -> Graph:
    g = Graph()
    if metadata.license_iri:
        g.add((file_bn, KNORA_API.hasLicense, URIRef(metadata.license_iri)))

    literal_metadata = []
    if metadata.copyright_holder:
        literal_metadata.append((KNORA_API.hasCopyrightHolder, metadata.copyright_holder))
    if metadata.authorships:
        for auth in metadata.authorships:
            literal_metadata.append((KNORA_API.hasAuthorship, auth))
    if metadata.permissions:
        literal_metadata.append((KNORA_API.hasPermissions, metadata.permissions))
    for prop, literal_val in literal_metadata:
        g.add((file_bn, prop, Literal(literal_val, datatype=XSD.string)))
    return g



from typing import Any
from typing import Optional
from typing import Union

from loguru import logger

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.clients.authentication_client_live import AuthenticationClientLive
from dsp_tools.clients.connection import Connection
from dsp_tools.clients.connection_live import ConnectionLive
from dsp_tools.commands.project.create.project_validate import validate_project
from dsp_tools.commands.project.legacy_models.listnode import ListNode
from dsp_tools.commands.project.legacy_models.project import Project
from dsp_tools.error.exceptions import BaseError
from dsp_tools.error.exceptions import InputError
from dsp_tools.utils.json_parsing import parse_json_input


def create_lists_on_server(
    lists_to_create: list[dict[str, Any]],
    con: Connection,
    project_remote: Project,
) -> tuple[dict[str, Any], bool]:
    """
    Creates the "lists" section of a JSON project definition on a DSP server.
    If a list with the same name is already existing in this project on the DSP server, this list is skipped.
    If a node or an entire list cannot be created, an error message is printed, but the process continues.

    Args:
        lists_to_create: "lists" section of a JSON project definition
        con: connection to the DSP server
        project_remote: representation of the project on the DSP server

    Raises:
        BaseError: if one of the lists to be created already exists on the DSP server, but it has no name

    Returns:
        tuple consisting of the IRIs of the list nodes and the success status (True if everything went well)
    """

    overall_success = True

    # retrieve existing lists
    try:
        existing_lists = ListNode.getAllLists(con=con, project_iri=project_remote.iri)
    except BaseError:
        err_msg = "Unable to retrieve existing lists on DSP server. Cannot check if your lists are already existing."
        print(f"WARNING: {err_msg}")
        logger.exception(err_msg)
        existing_lists = []
        overall_success = False

    current_project_lists: dict[str, Any] = {}
    for new_lst in lists_to_create:
        if existing_lst := [x for x in existing_lists if x.project == project_remote.iri and x.name == new_lst["name"]]:
            existing_list_name = existing_lst[0].name
            if not existing_list_name:
                raise BaseError(f"Node {existing_lst[0]} has no name.")
            current_project_lists[existing_list_name] = {
                "id": existing_lst[0].iri,
                "nodes": new_lst["nodes"],
            }
            print(f"    WARNING: List '{new_lst['name']}' already exists on the DSP server. Skipping...")
            overall_success = False
            continue

        created_list, success = _create_list_node(con=con, project=project_remote, node=new_lst)
        current_project_lists.update(created_list)
        if not success:
            overall_success = False
        print(f"    Created list '{new_lst['name']}'.")

    return current_project_lists, overall_success


def _create_list_node(
    con: Connection,
    project: Project,
    node: dict[str, Any],
    parent_node: Optional[ListNode] = None,
) -> tuple[dict[str, Any], bool]:
    """
    Creates a list node on the DSP server, recursively scanning through all its subnodes, creating them as well.
    If a node cannot be created, an error message is printed, but the process continues.

    Args:
        con: connection to the DSP server
        project: project that holds the list where this node should be added to
        node: the node to be created
        parent_node: parent node of the node to be created (optional)

    Returns:
        Returns a tuple consisting of a dict and a bool.
        The dict contains the IRIs of the created list nodes,
        nested according to their hierarchy structure,
        i.e. ``{nodename: {"id": IRI, "nodes": {...}}}``.
        The bool is True if all nodes could be created,
        False if any node could not be created.

    Raises:
        BaseError: if the created node has no name
    """
    new_node: ListNode = ListNode(
        con=con,
        project=project,
        label=node["labels"],
        comments=node.get("comments"),
        name=node["name"],
        parent=parent_node,
    )
    try:
        new_node = new_node.create()
    except BaseError:
        print(f"WARNING: Cannot create list node '{node['name']}'.")
        logger.exception(f"Cannot create list node '{node['name']}'.")
        return {}, False

    # if node has child nodes, call the method recursively
    if node.get("nodes"):
        overall_success = True
        subnode_list = []
        for subnode in node["nodes"]:
            created_subnode, success = _create_list_node(con=con, project=project, node=subnode, parent_node=new_node)
            subnode_list.append(created_subnode)
            if not success:
                overall_success = False
        if not new_node.name:
            raise BaseError(f"Node {new_node} has no name.")
        return {new_node.name: {"id": new_node.iri, "nodes": subnode_list}}, overall_success
    else:
        if not new_node.name:
            raise BaseError(f"Node {new_node} has no name.")
        return {new_node.name: {"id": new_node.iri}}, True


def create_only_lists(
    project_file_as_path_or_parsed: Union[str, dict[str, Any]],
    creds: ServerCredentials,
) -> tuple[dict[str, Any], bool]:
    """
    This function accepts a JSON project definition,
    connects to a DSP server,
    and uploads the "lists" section to the server.

    The project must already exist on the DSP server.
    If a list with the same name is already existing in this project on the DSP server, this list is skipped.

    Args:
        project_file_as_path_or_parsed: path to the JSON project definition, or parsed JSON object
        creds: credentials to connect to the DSP server

    Raises:
        InputError:
          - if the project cannot be read from the server
          - if the connection to the DSP server cannot be established

        BaseError:
          - if the input is invalid
          - if a problem occurred while trying to expand the Excel files
          - if the JSON file is invalid according to the schema

    Returns:
        Returns a tuple consisting of a dict and a bool.
        The dict contains the IRIs of the created list nodes,
        nested according to their hierarchy structure,
        i.e. ``{nodename: {"id": IRI, "nodes": {...}}}``.
        If there are no lists in the project definition,
        an empty dictionary is returned.
        The bool indicates if everything went smoothly during the process.
        If a warning or error occurred (e.g. one of the lists already exists,
        or one of the nodes could not be created),
        it is False.
    """
    project_definition = parse_json_input(project_file_as_path_or_parsed=project_file_as_path_or_parsed)
    if not project_definition.get("project", {}).get("lists"):
        return {}, True
    validate_project(project_definition)
    print("JSON project file is syntactically correct and passed validation.")

    auth = AuthenticationClientLive(creds.server, creds.user, creds.password)
    con = ConnectionLive(creds.server, auth)

    # retrieve the project
    shortcode = project_definition["project"]["shortcode"]
    project_local = Project(con=con, shortcode=shortcode)
    try:
        project_remote = project_local.read()
    except BaseError:
        err_msg = f"Unable to create the lists: The project {shortcode} cannot be found on the DSP server."
        logger.exception(err_msg)
        raise InputError(err_msg) from None

    # create new lists
    current_project_lists, success = create_lists_on_server(
        lists_to_create=project_definition["project"]["lists"],
        con=con,
        project_remote=project_remote,
    )

    return current_project_lists, success



# The accepted XML tags are defined at https://docs.dasch.swiss/latest/DSP-API/03-endpoints/api-v2/text/standard-standoff/
_COMMON_BASE = [
    "p",
    "em",
    "strong",
    "u",
    "sub",
    "sup",
    "strike",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "ol",
    "ul",
    "li",
    "tbody",
    "table",
    "tr",
    "td",
    "br",
    "hr",
    "pre",
    "cite",
    "blockquote",
    "code",
]
KNOWN_XML_TAG_REGEXES = [
    *_COMMON_BASE,
    "a( [^>]+)?",  # <a> has attributes
    "footnote( [^>]+)?",  # the footnote text is in the attributes
]
KNOWN_XML_TAGS = [
    *_COMMON_BASE,
    "a",
    "footnote",
]
XML_NAMESPACE_MAP = {None: "https://dasch.swiss/schema", "xsi": "http://www.w3.org/2001/XMLSchema-instance"}
DASCH_SCHEMA = "{https://dasch.swiss/schema}"



from __future__ import annotations

from pathlib import Path

from loguru import logger
from lxml import etree

from dsp_tools.error.exceptions import InputError


def parse_xml_file(input_file: str | Path) -> etree._Element:
    """
    This function parses an XML file and returns an Element Tree

    Args:
        input_file: path to the input file

    Returns:
        element tree

    Raises:
        InputError: if the file contains a syntax error
    """
    parser = etree.XMLParser(remove_comments=True, remove_pis=True)
    try:
        return etree.parse(source=input_file, parser=parser).getroot()
    except etree.XMLSyntaxError as err:
        logger.error(f"The XML file contains the following syntax error: {err.msg}")
        raise InputError(f"The XML file contains the following syntax error: {err.msg}") from None



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Literal

from rdflib import URIRef

from dsp_tools.commands.xmlupload.iri_resolver import IriResolver
from dsp_tools.commands.xmlupload.models.permission import Permissions


@dataclass
class IntermediaryLookups:
    permissions: dict[str, Permissions]
    listnodes: dict[tuple[str, str], str]
    namespaces: dict[str, str]
    authorships: dict[str, list[str]]


@dataclass
class IRILookups:
    project_iri: URIRef
    id_to_iri: IriResolver


@dataclass(frozen=True)
class OntoInfo:
    """This class saves the properties and the classes from an ontology."""

    classes: list[str] = field(default_factory=list)
    properties: list[str] = field(default_factory=list)


@dataclass
class ProjectOntosInformation:
    """This class saves information needed to check the consistency with the ontology."""

    default_ontology_prefix: str
    onto_lookup: dict[str, OntoInfo]


def make_project_onto_information(
    default_onto_prefix: str,
    ontologies: dict[str, list[dict[str, Any]]],
) -> ProjectOntosInformation:
    """
    This function formats ontologies returned by the dsp-api into a look-up model.

    Args:
        default_onto_prefix: prefix for the default ontology
        ontologies: ontologies returned from the dsp-api

    Returns:
        A look-up that contains the property and class names from each ontology
    """
    onto_dict = {
        onto_name: _extract_classes_and_properties_from_onto(onto_json) for onto_name, onto_json in ontologies.items()
    }
    return ProjectOntosInformation(default_onto_prefix, onto_dict)


def _extract_classes_and_properties_from_onto(onto_json: list[dict[str, Any]]) -> OntoInfo:
    """
    This function takes an ontology response from the DSP-API.
    It extracts the classes and properties.
    And saves them in an instance of the class Ontology.

    Args:
        onto_json: response from DSP-API

    Returns:
        Ontology instance with the classes and properties
    """
    classes = _get_all_cleaned_classes_from_onto(onto_json)
    properties = _get_all_cleaned_properties_from_onto(onto_json)
    return OntoInfo(classes, properties)


def _get_all_cleaned_classes_from_onto(onto_json: list[dict[str, Any]]) -> list[str]:
    classes = _get_all_classes_from_onto(onto_json)
    return _remove_prefixes(classes)


def _get_all_classes_from_onto(onto_json: list[dict[str, Any]]) -> list[str]:
    return [elem["@id"] for elem in onto_json if elem.get("knora-api:isResourceClass")]


def _get_all_cleaned_properties_from_onto(onto_json: list[dict[str, Any]]) -> list[str]:
    props = _get_all_properties_from_onto(onto_json)
    return _remove_prefixes(props)


def _get_all_properties_from_onto(onto_json: list[dict[str, Any]]) -> list[str]:
    return [elem["@id"] for elem in onto_json if not elem.get("knora-api:isResourceClass")]


def _remove_prefixes(ontology_elements: list[str]) -> list[str]:
    return [x.split(":")[1] for x in ontology_elements]


AllowedEncodings = Literal["utf8", "xml"]


@dataclass
class TextValueData:
    resource_id: str
    res_type: str
    property_name: str
    encoding: AllowedEncodings


@dataclass
class PropertyTextValueTypes:
    """
    This class contains the information
    which properties have which type of encoding for a TextValue property in the ontology.
    """

    formatted_text_props: set[str] = field(default_factory=set)
    unformatted_text_props: set[str] = field(default_factory=set)


def get_text_value_types_of_properties_from_onto(
    onto_json_dict: dict[str, list[dict[str, Any]]],
) -> PropertyTextValueTypes:
    """
    This function takes a dict with the project ontologies in the format:
        { ontology_name: ontology_json_from_api }
    It retrieves the properties that are used with `knora-api:TextValue`.
    They are separated into two categories: xml encoded ones, and utf8 encoded ones.

    Args:
        onto_json_dict: dict with the project ontologies

    Returns:
        Look-up containing the properties separated according to the formatting
    """
    all_props = []
    for onto_json in onto_json_dict.values():
        all_props.extend(_get_all_text_value_types_properties_and_from_onto(onto_json))
    return _make_text_value_property_type_lookup(all_props)


def _make_text_value_property_type_lookup(prop_list: list[tuple[str, str]]) -> PropertyTextValueTypes:
    formatted_text = {p for p, _type in prop_list if _type == "salsah-gui:Richtext"}
    formatted_text.update(["hasComment", "hasDescription"])  # knora-api properties that can be used directly
    unformatted_text = {p for p, _type in prop_list if _type != "salsah-gui:Richtext"}
    unformatted_text.update(["hasTitle", "hasKeyword"])  # knora-api properties that can be used directly
    return PropertyTextValueTypes(formatted_text, unformatted_text)


def _get_all_text_value_types_properties_and_from_onto(onto_json: list[dict[str, Any]]) -> list[tuple[str, str]]:
    prop_id_list = [elem["@id"] for elem in onto_json if _check_if_text_value_property(elem)]
    type_list = [elem["salsah-gui:guiElement"]["@id"] for elem in onto_json if _check_if_text_value_property(elem)]
    return list(zip(prop_id_list, type_list))


def _check_if_text_value_property(onto_ele: dict[str, Any]) -> bool:
    if not onto_ele.get("knora-api:isResourceProperty"):
        return False
    if not (object_type := onto_ele.get("knora-api:objectType")):
        return False
    match object_type["@id"]:
        case "knora-api:TextValue":
            return True
        case _:
            return False


def make_namespace_dict_from_onto_names(ontos: dict[str, str]) -> dict[str, str]:
    """Provided a dictionary of ontology names and IRIs, returns a dictionary of Namespace objects."""
    ontos = _correct_project_context_namespaces(ontos)
    ontos["knora-api"] = "http://api.knora.org/ontology/knora-api/v2#"
    return ontos


def _correct_project_context_namespaces(ontos: dict[str, str]) -> dict[str, str]:
    """Add the hashtag to make it a valid namespace."""
    return {k: f"{v}#" for k, v in ontos.items()}



from typing import Protocol
from typing import runtime_checkable

from dsp_tools.utils.ansi_colors import BOLD_RED
from dsp_tools.utils.ansi_colors import RESET_TO_DEFAULT
from dsp_tools.utils.ansi_colors import YELLOW


@runtime_checkable
class DspToolsWarning(Protocol):
    """Protocol for warnings that implement a custom showwarnings() function"""

    @classmethod
    def showwarning(cls, message: str) -> None:
        """Functionality that should be executed when a warning of this class is emitted"""


class DspToolsUserWarning(Warning):
    """Class for general user-facing warnings"""

    @classmethod
    def showwarning(cls, message: str) -> None:
        """Print the warning, without context"""
        print(BOLD_RED + f"WARNING: {message}" + RESET_TO_DEFAULT)


class DspToolsUserInfo(Warning):
    """Class for general user-facing warnings"""

    @classmethod
    def showwarning(cls, message: str) -> None:
        """Print the information, without context"""
        print(YELLOW + f"INFO: {message}" + RESET_TO_DEFAULT)


class DspToolsFutureWarning(FutureWarning):
    """Class for user-facing deprecation warnings"""

    @classmethod
    def showwarning(cls, message: str) -> None:
        """Print the warning, without context"""
        print(BOLD_RED + f"DEPRECATION WARNING: {message}" + RESET_TO_DEFAULT)



import warnings

import pytest

from dsp_tools.error.custom_warnings import DspToolsUserWarning
from dsp_tools.xmllib.models.config_options import Permissions
from dsp_tools.xmllib.models.values import BooleanValue
from dsp_tools.xmllib.models.values import ColorValue
from dsp_tools.xmllib.models.values import DateValue
from dsp_tools.xmllib.models.values import DecimalValue
from dsp_tools.xmllib.models.values import GeonameValue
from dsp_tools.xmllib.models.values import IntValue
from dsp_tools.xmllib.models.values import LinkValue
from dsp_tools.xmllib.models.values import ListValue
from dsp_tools.xmllib.models.values import Richtext
from dsp_tools.xmllib.models.values import SimpleText
from dsp_tools.xmllib.models.values import TimeValue
from dsp_tools.xmllib.models.values import UriValue


class TestBooleanValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            BooleanValue("False", ":booleanProp", resource_id="res_id", permissions=Permissions.OPEN)
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            BooleanValue("other", ":booleanProp", resource_id="res_id", permissions=Permissions.OPEN)


class TestColorValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            ColorValue("#FFFFFF", ":colorProp", resource_id="res_id", permissions=Permissions.OPEN)
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            ColorValue("invalidColor", ":colorProp", resource_id="res_id", permissions=Permissions.OPEN)


class TestDateValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            DateValue("2023-01-01", ":dateProp", resource_id="res_id", permissions=Permissions.OPEN)
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            DateValue("invalidDate", ":dateProp", resource_id="res_id", permissions=Permissions.OPEN)


class TestDecimalValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            DecimalValue("3.14", ":decimalProp", resource_id="res_id", permissions=Permissions.OPEN)
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            DecimalValue("invalidDecimal", ":decimalProp", resource_id="res_id", permissions=Permissions.OPEN)


class TestGeonameValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            GeonameValue("00099", ":geonameProp", resource_id="res_id", permissions=Permissions.OPEN)
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            GeonameValue("invalidGeoname", ":geonameProp", resource_id="res_id", permissions=Permissions.OPEN)


class TestIntValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            IntValue("42", ":intProp", resource_id="res_id", permissions=Permissions.OPEN)
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            IntValue("invalidInt", ":intProp", resource_id="res_id", permissions=Permissions.OPEN)


class TestLinkValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            LinkValue("link", ":linkProp", resource_id="res_id", permissions=Permissions.OPEN)
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            LinkValue(None, ":linkProp", resource_id="res_id", permissions=Permissions.OPEN)  # type: ignore[arg-type]


class TestListValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            ListValue("item1", "listName", ":listProp", resource_id="res_id", permissions=Permissions.OPEN)
        assert len(caught_warnings) == 0

    def test_warns_false_node(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            ListValue(None, "list", ":listProp", resource_id="res_id", permissions=Permissions.OPEN)

    def test_warns_false_list(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            ListValue("Node", None, ":listProp", resource_id="res_id", permissions=Permissions.OPEN)


class TestRichtext:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            Richtext("<p>Hello World</p>", ":richtextProp", resource_id="res_id", permissions=Permissions.OPEN)
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            Richtext(None, ":richtextProp", resource_id="res_id", permissions=Permissions.OPEN)  # type: ignore[arg-type]


class TestSimpleText:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            SimpleText("Hello World", ":simpleTextProp", resource_id="res_id", permissions=Permissions.OPEN)
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            SimpleText(None, ":simpleTextProp", resource_id="res_id", permissions=Permissions.OPEN)  # type: ignore[arg-type]


class TestTimeValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            TimeValue("2009-10-10T12:00:00-05:00", ":timeProp", resource_id="res_id", permissions=Permissions.OPEN)
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            TimeValue("invalidTime", ":timeProp", resource_id="res_id", permissions=Permissions.OPEN)


class TestUriValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            UriValue("https://example.com", ":uriProp", resource_id="res_id", permissions=Permissions.OPEN)
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            UriValue("invalidUri", ":uriProp", resource_id="res_id", permissions=Permissions.OPEN)


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from dataclasses import dataclass

from rdflib import Graph
from rdflib import URIRef


@dataclass
class SHACLValidationReport:
    conforms: bool
    validation_graph: Graph


@dataclass
class AllProjectLists:
    all_lists: list[OneList]


@dataclass
class OneList:
    list_iri: str
    list_name: str
    nodes: list[str]

    def hlist(self) -> str:
        return f'"hlist=<{self.list_iri}>"'


@dataclass
class SHACLListInfo:
    list_iri: URIRef
    sh_path: URIRef
    sh_message: str
    sh_in: list[str]



import unittest

import jsonpath_ng
import pytest
import regex

from dsp_tools.commands.excel2json import properties as e2j
from dsp_tools.error.exceptions import InputError

excelfile = "testdata/excel2json/old_excel2json_files/test-name (test_label)/properties.xlsx"
output_from_method, _ = e2j.excel2properties(excelfile, None)


# ruff: noqa: PT009 (pytest-unittest-assertion) (remove this line when pytest is used instead of unittest)
class TestValidateProperties:
    # it is not possible to call the method to be tested directly.
    # So let's make a reference to it, so that it can be found by the usage search
    lambda _: e2j._validate_properties_section_in_json([])

    def test_invalid_super(self) -> None:
        expected_msg = (
            "The Excel file 'properties.xlsx' contains the following problems:\n\n\n"
            "    Section of the problem: 'Properties'\n"
            "    Problematic property: 'hasGeoname'\n"
            "    Located at: Column 'super' | Row 3\n"
            "    Original Error Message: 'GeonameValue' is not valid under any of the given schemas"
        )
        with pytest.raises(InputError, match=expected_msg):
            e2j.excel2properties(
                excelfile="testdata/invalid-testdata/excel2json/properties-invalid-super.xlsx", path_to_output_file=""
            )

    def test_invalid_object(self) -> None:
        expected_msg = regex.escape(
            "The Excel file 'properties.xlsx' contains the following problems:\n\n\n"
            "    Section of the problem: 'Properties'\n"
            "    Problematic property: 'hasBoolean'\n"
            "    Located at: Column 'object' | Row 2\n"
            "    Original Error Message: 'hasValue' is not valid under any of the given schemas"
        )
        with pytest.raises(InputError, match=expected_msg):
            e2j.excel2properties(
                excelfile="testdata/invalid-testdata/excel2json/properties-invalid-object.xlsx", path_to_output_file=""
            )

    def test_invalid_gui_element(self) -> None:
        expected_msg = regex.escape(
            "The Excel file 'properties.xlsx' contains the following problems:\n\n\n"
            "    Section of the problem: 'Properties'\n"
            "    Problematic property: 'hasGeoname'\n"
            "    Located at: Column 'gui_element' | Row 3\n"
            "    Original Error Message: 'Geonames' was expected"
        )
        with pytest.raises(InputError, match=expected_msg):
            e2j.excel2properties(
                excelfile="testdata/invalid-testdata/excel2json/properties-invalid-gui_element.xlsx",
                path_to_output_file="",
            )

    def test_invalid_gui_attrib_values(self) -> None:
        expected_msg = regex.escape(
            "The Excel file 'properties.xlsx' contains the following problems:\n\n\n"
            "    Section of the problem: 'Properties'\n"
            "    Problematic property: 'hasInteger'\n"
            "    Located at: Column 'gui_attributes' | Row 4\n"
            "    Original Error Message: Additional properties are not allowed ('rows' was unexpected)"
        )
        with pytest.raises(InputError, match=expected_msg):
            e2j.excel2properties(
                excelfile="testdata/invalid-testdata/excel2json/properties-invalid-gui_attribute_values.xlsx",
                path_to_output_file="",
            )


def test_excel2properties_invalid_gui_attrib_format() -> None:
    expected_msg = regex.escape(
        "The Excel file 'properties.xlsx' contains the following problems:\n\n"
        "The property 'hasInteger' has the following problem(s):\n"
        "There is invalid content in the excel.\n"
        "Located at: Column 'gui_attributes' | Row 4\n"
        "Expected Content: attribute1: value, attribute2: value (no attribute key may be duplicated)\n"
        "Actual Content: max=10, min=5"
    )
    with pytest.raises(InputError, match=expected_msg):
        e2j.excel2properties("testdata/invalid-testdata/excel2json/properties-invalid-gui_attribute_format.xlsx", "")


class TestExcelToProperties(unittest.TestCase):
    def test_names(self) -> None:
        excel_names = [
            "correspondsToGenericAnthroponym",
            "hasAnthroponym",
            "hasGender",
            "isDesignatedAs",
            "hasTitle",
            "hasStatus",
            "hasLifeYearAmount",
            "hasBirthDate",
            "hasRepresentation",
            "hasRemarks",
            "hasTerminusPostQuem",
            "hasGND",
            "hasColor",
            "hasDecimal",
            "hasTime",
            "hasBoolean",
            "hasGeoname",
            "partOfDocument",
            "linkstoRegion",
            "hasLinkToImage",
            "hasLinkToResource",
            "hasLinkToArchiveRepresentation",
            "hasLinkToMovingImageRepesentation",
            "hasLinkToAudioRepesentation",
        ]
        json_names = [match.value for match in jsonpath_ng.parse("$[*].name").find(output_from_method)]
        self.assertListEqual(excel_names, json_names)

    def test_supers(self) -> None:
        excel_supers = [
            ["hasLinkTo"],
            ["hasValue", "dcterms:creator"],
            ["hasValue"],
            ["hasValue"],
            ["hasLinkTo"],
            ["hasValue"],
            ["hasValue"],
            ["hasValue"],
            ["hasRepresentation"],
            ["hasValue", "dcterms:description"],
            ["hasValue"],
            ["hasValue"],
            ["hasColor"],
            ["hasValue"],
            ["hasValue"],
            ["hasValue"],
            ["hasValue"],
            ["isPartOf"],
            ["hasLinkTo"],
            ["hasLinkTo"],
            ["hasLinkTo"],
            ["hasLinkTo"],
            ["hasLinkTo"],
            ["hasLinkTo"],
        ]
        json_supers = [match.value for match in jsonpath_ng.parse("$[*].super").find(output_from_method)]
        self.assertListEqual(excel_supers, json_supers)

    def test_objects(self) -> None:
        excel_objects = [
            ":GenericAnthroponym",
            "TextValue",
            "ListValue",
            "ListValue",
            ":Titles",
            "ListValue",
            "IntValue",
            "DateValue",
            "Representation",
            "TextValue",
            "DateValue",
            "UriValue",
            "ColorValue",
            "DecimalValue",
            "TimeValue",
            "BooleanValue",
            "GeonameValue",
            ":Documents",
            "Region",
            "StillImageRepresentation",
            "Resource",
            "ArchiveRepresentation",
            "MovingImageRepresentation",
            "AudioRepresentation",
        ]
        json_objects = [match.value for match in jsonpath_ng.parse("$[*].object").find(output_from_method)]
        self.assertListEqual(excel_objects, json_objects)

    def test_labels(self) -> None:
        excel_labels = {
            "de": [
                "",
                "only German",
                "",
                "",
                "",
                "",
                "",
                "",
                "hat eine Multimediadatei",
                "",
                "",
                "GND",
                "Farbe",
                "Dezimalzahl",
                "Zeit",
                "Bool'sche Variable",
                "Link zu Geonames",
                "ist Teil eines Dokuments",
                "",
                "",
                "",
                "",
                "",
                "",
            ],
            "it": [
                "",
                "",
                "",
                "only Italian",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "GND",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
            ],
        }  # there are also labels in other languages, but they are not tested
        json_labels_all = [match.value for match in jsonpath_ng.parse("$[*].labels").find(output_from_method)]
        json_labels = {lang: [label.get(lang, "").strip() for label in json_labels_all] for lang in ["de", "it"]}
        self.assertDictEqual(excel_labels, json_labels)

    def test_comments(self) -> None:
        excel_comments = {
            "comment_fr": [
                "J'avais déjà examiné plusieurs propriétés quand, un jour, le notaire, qui me "
                "donnait des indications nécessaires pour une de mes explorations, me dit :",
                "Un étrange hasard m'a mis en possession de ce journal.",
                "Je n'en sais rien du tout ; mais si vous voulez la voir, monsieur, voici les "
                "indications précises pour la trouver.",
                "Vous devrez arranger l'affaire avec le curé du village de --.\"",
                "Un étrange hasard m'a mis en possession de ce journal.",
                "",
                "",
                "only French",
                "",
                "",
                "J'avais déjà examiné plusieurs propriétés quand, un jour, le notaire, qui me "
                "donnait des indications nécessaires pour une de mes explorations, me dit :",
                "Gemeinsame Normdatei",
                "",
                "Chiffre décimale",
                "Temps",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
            ],
            "comment_it": [
                "Avevo già visto diverse proprietà quando un giorno il notaio,",
                "Uno strano caso mi mise in possesso di questo diario.",
                "Non ne so nulla; ma se volete vederla, signore, eccovi le indicazioni precise per trovarla.",
                "Dovrete organizzare l'affare con il curato del villaggio di --\".",
                "Uno strano caso mi mise in possesso di questo diario.",
                "",
                "",
                "",
                "",
                "",
                "Avevo già visto diverse proprietà quando un giorno il notaio,",
                "Gemeinsame Normdatei",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
            ],
        }  # there are also comments in other languages, but they are not tested
        json_comments = {
            f"comment_{lang}": [resource.get("comments", {}).get(lang, "").strip() for resource in output_from_method]
            for lang in ["fr", "it"]
        }
        self.assertDictEqual(excel_comments, json_comments)

    def test_gui_elements(self) -> None:
        excel_gui_elements = [
            "Searchbox",
            "Richtext",
            "List",
            "Radio",
            "Searchbox",
            "List",
            "Spinbox",
            "Date",
            "Searchbox",
            "Textarea",
            "Date",
            "SimpleText",
            "Colorpicker",
            "Spinbox",
            "TimeStamp",
            "Checkbox",
            "Geonames",
            "Searchbox",
            "Searchbox",
            "Searchbox",
            "Searchbox",
            "Searchbox",
            "Searchbox",
            "Searchbox",
        ]

        json_gui_elements = [match.value for match in jsonpath_ng.parse("$[*].gui_element").find(output_from_method)]
        self.assertListEqual(excel_gui_elements, json_gui_elements)

    def test_gui_attributes_hasGender(self) -> None:
        excel_gui_attributes_hasGender = {"hlist": "gender"}
        json_gui_attributes_hasGender = (
            jsonpath_ng.ext.parse("$[?name='hasGender'].gui_attributes").find(output_from_method)[0].value
        )
        self.assertDictEqual(excel_gui_attributes_hasGender, json_gui_attributes_hasGender)

    def test_gui_attributes_hasGND(self) -> None:
        excel_gui_attributes_hasGND = {"size": 100}
        json_gui_attributes_hasGND = (
            jsonpath_ng.ext.parse("$[?name='hasGND'].gui_attributes").find(output_from_method)[0].value
        )
        self.assertDictEqual(excel_gui_attributes_hasGND, json_gui_attributes_hasGND)

    def test_gui_attributes_hasDecimal(self) -> None:
        excel_gui_attributes_hasDecimal = {"min": 0.0, "max": 100.0}
        json_gui_attributes_hasDecimal = (
            jsonpath_ng.ext.parse("$[?name='hasDecimal'].gui_attributes").find(output_from_method)[0].value
        )
        self.assertDictEqual(excel_gui_attributes_hasDecimal, json_gui_attributes_hasDecimal)


if __name__ == "__main__":
    pytest.main([__file__])



from enum import Enum
from enum import unique
from typing import Any
from typing import Optional
from typing import Union

from dsp_tools.error.exceptions import BaseError


@unique
class Languages(Enum):
    """Languages supported by DSP"""

    EN = "en"
    DE = "de"
    FR = "fr"
    IT = "it"
    RM = "rm"


LangStringParam = Optional[Union[dict[Union[Languages, str], str], str]]


class LangStringIterator:
    """Iterator class for LangString class."""

    _langstring: "LangString"
    _index: int

    def __init__(self, langstring: "LangString"):
        self._langstring = langstring
        self._langlist = list(map(lambda a: a[0], self._langstring.items()))
        self._index = 0

    def __next__(self):
        if len(self._langlist) == 0 and self._index == 0:
            return None, self._langstring[None]
        elif self._index < len(self._langlist):
            tmp = self._langlist[self._index]
            self._index += 1
            return tmp, self._langstring[tmp]
        else:
            raise StopIteration


class LangString:
    """
    This class holds a typical language-string pair as it is often used in JSON-LD as
    ```
    "some:thing" = [{
      '@language': 'xx',
      '@value': 'a string in language xx'
    },
    {…},…]
    ```

    or a simple string without language:
    ```
    "some:thing": "a string without language specificer"
    ```
    """

    _langstrs: dict[Languages, str]
    _simplestring: str

    def __init__(self, initvalue: LangStringParam = None):
        def mymapper(p: tuple[Union[Languages, str], str]) -> tuple[Languages, str]:
            lmap = dict(map(lambda a: (a.value, a), Languages))
            if isinstance(p[0], str) and p[0] in lmap:
                lang = lmap[p[0].lower()]
            elif isinstance(p[0], Languages):
                lang = p[0]
            else:
                raise BaseError("Not a valid language definition!")
            return lang, p[1]

        if initvalue is None:
            self._simplestring = None
            self._langstrs = {}
        elif isinstance(initvalue, str):
            self._simplestring = initvalue
            self._langstrs = {}
        elif isinstance(initvalue, dict):
            self._simplestring = None
            self._langstrs = dict(map(mymapper, initvalue.items()))
        elif isinstance(initvalue, LangString):
            self._simplestring = initvalue._simplestring
            self._langstrs = initvalue._langstrs
        else:
            raise BaseError("Not a valid language definition!")

    def __getitem__(self, key: Optional[Union[Languages, str]] = None) -> Optional[str]:
        match key:
            case None:
                return self.__getitem_nokey__()
            case Enum() if ls := self._langstrs.get(key):
                return ls
            case Enum():
                return self.__getitem_fallback__()
            case str():
                lmap = {a.value: a for a in Languages}
                lkey = lmap.get(key.lower())
                if lkey is None:
                    raise BaseError(f"Invalid language string '{key}'")
                if ls := self._langstrs.get(lkey):
                    return ls
                return self.__getitem_fallback__()

    def __getitem_nokey__(self) -> Optional[str]:
        if self._simplestring:
            return self._simplestring
        elif self._langstrs:
            return next(iter(self._langstrs)).value
        else:
            return None

    def __getitem_fallback__(self) -> Optional[str]:
        for lst in self._langstrs:
            if self._langstrs.get(lst) is not None:
                return self._langstrs[lst]
        if self._simplestring is not None:
            return self._simplestring
        return None

    def __setitem__(self, key: Optional[Union[Languages, str]], value: str):
        if key is None:
            self._simplestring = value
            self._langstrs = {}  # Delete language dependent string! Needs Caution!!!
            return
        if isinstance(key, Languages):
            self._langstrs[key] = value
        else:
            lmap = dict(map(lambda a: (a.value, a), Languages))
            if lmap.get(key.lower()) is None:
                raise BaseError('Invalid language string "' + key + '"!')
            self._langstrs[lmap[key.lower()]] = value

    def __delitem__(self, key: Union[Languages, str]):
        if isinstance(key, Languages):
            del self._langstrs[key]
        else:
            lmap = dict(map(lambda a: (a.value, a), Languages))
            if lmap.get(key.lower()) is None:
                raise BaseError('Invalid language string "' + key + '"!')
            del self._langstrs[lmap[key.lower()]]

    def __str__(self):
        tmpstr = "{"
        for p in self._langstrs:
            tmpstr += "=" + p.value + ":" + self._langstrs[p]
        tmpstr += "}"
        return tmpstr

    def __iter__(self):
        return LangStringIterator(self)

    def __eq__(self, other: "LangString"):
        equal = self._simplestring == other._simplestring
        if equal:
            for lang in Languages:
                equal = self._langstrs.get(lang) == other._langstrs.get(lang)
                if not equal:
                    break
        return equal

    def items(self):
        return self._langstrs.items()

    def isEmpty(self) -> bool:
        return not (bool(self._langstrs) or bool(self._simplestring))

    def empty(self) -> None:
        self._simplestring = None
        self._langstrs = {}

    def toJsonObj(self) -> Optional[Union[str, list[dict[str, str]]]]:
        if self.isEmpty():
            return None
        if self._simplestring is not None:
            return self._simplestring
        else:
            return list(map(lambda a: {"language": a[0].value, "value": a[1] if a[1] else "-"}, self._langstrs.items()))

    def toJsonLdObj(self) -> Optional[Union[str, list[dict[str, str]]]]:
        if self.isEmpty():
            return None
        if self._simplestring is not None:
            return self._simplestring
        else:
            return [{"@language": a[0].value, "@value": a[1]} for a in self._langstrs.items()]
            # return list(map(lambda a: {'@language': a[0].value, '@value': a[1]}, self._langstrs.items()))

    @classmethod
    def fromJsonLdObj(cls, obj: Optional[Union[list[dict[str, str]], str]]) -> "LangString":
        if obj is None:
            return None
        if isinstance(obj, str):
            return cls(obj)
        if isinstance(obj, list):
            objs = obj
        else:
            objs = [obj]
        lstrs: dict[Languages, str] = {}
        for o in objs:
            lang = o.get("@language")
            if lang == "en":
                lstrs[Languages.EN] = o.get("@value")
            elif lang == "de":
                lstrs[Languages.DE] = o.get("@value")
            elif lang == "fr":
                lstrs[Languages.FR] = o.get("@value")
            elif lang == "it":
                lstrs[Languages.IT] = o.get("@value")
            elif lang == "rm":
                lstrs[Languages.RM] = o.get("@value")
            elif o.get("@value") is not None:
                return cls(o.get("@value"))
        return cls(lstrs)

    @classmethod
    def fromJsonObj(cls, obj: Optional[Any]) -> "LangString":
        if obj is None:
            return None
        if isinstance(obj, str):
            return cls(obj)
        if isinstance(obj, list):
            objs = obj
        else:
            objs = [obj]
        lstrs: dict[Languages, str] = {}
        for o in objs:
            lang = o.get("language")
            if lang == "en":
                lstrs[Languages.EN] = o.get("value")
            elif lang == "de":
                lstrs[Languages.DE] = o.get("value")
            elif lang == "fr":
                lstrs[Languages.FR] = o.get("value")
            elif lang == "it":
                lstrs[Languages.IT] = o.get("value")
            elif lang == "rm":
                lstrs[Languages.RM] = o.get("value")
            elif o.get("value") is not None:
                return cls(o.get("value"))
        return cls(lstrs)

    def createDefinitionFileObj(self) -> Union[str, dict[str, str]]:
        if self._simplestring:
            return self._simplestring
        langstring = {}
        for p in self.items():
            langstring[p[0].value] = p[1]
        return langstring
